import types

import pytest
from unittest.mock import Mock

from conan import ConanFile
from conan.internal.default_settings import default_settings_yml
from conan.tools.cmake import CMakeToolchain
from conan.tools.cmake.toolchain.blocks import Block
from conan.errors import ConanException
from conan.internal.model.conf import Conf
from conan.internal.model.options import Options
from conan.internal.model.settings import Settings


@pytest.fixture
def base_conanfile():
    c = ConanFile()
    settings = Settings.loads(default_settings_yml)
    c.settings = settings
    c.settings.build_type = "Release"
    c.settings.arch = "x86"
    c.settings.compiler = "clang"
    c.settings.compiler.libcxx = "libstdc++"
    c.settings_build = c.settings
    c.settings.os = "Windows"
    c.conf = Conf()
    c.folders.set_base_generators(".")  # non-existing to not relativize
    conan_node = Mock()
    conan_node.dependencies = []
    conan_node.transitive_deps = {}
    conan_node.replaced_requires = {}
    c._conan_node = conan_node
    return c


@pytest.fixture
def conanfile(base_conanfile):
    c = base_conanfile
    c.conf.define("tools.cmake.cmaketoolchain:system_name", "potato")
    return c


class TestGenericBlocks:
    """ Tests general toolchain blocks operations, enabling, disabling, customizing
    """
    def test_remove(self, conanfile):
        toolchain = CMakeToolchain(conanfile)
        toolchain.blocks.remove("generic_system")
        content = toolchain.content
        assert 'CMAKE_SYSTEM_NAME' not in content
        assert "CMAKE_CXX_FLAGS_INIT" in content
        assert "_CMAKE_IN_TRY_COMPILE" in content

        # remove multiple
        toolchain = CMakeToolchain(conanfile)
        toolchain.blocks.remove("generic_system", "cmake_flags_init")
        content = toolchain.content
        assert 'CMAKE_SYSTEM_NAME' not in content
        assert "CMAKE_CXX_FLAGS_INIT" not in content
        assert "_CMAKE_IN_TRY_COMPILE" in content

    def test_select_blocks(self, conanfile):
        toolchain = CMakeToolchain(conanfile)
        toolchain.blocks.select("generic_system")
        content = toolchain.content
        assert "########## 'generic_system' block #############" in content
        assert "########## 'cmake_flags_init' block #############" not in content
        assert "########## 'libcxx' block #############" not in content
        # These are not removed by default, to not break behavior
        assert "########## 'variables' block #############" in content
        assert "########## 'preprocessor' block #############" in content
        assert 'CMAKE_SYSTEM_NAME' in content
        assert "CMAKE_CXX_FLAGS_INIT" not in content
        assert "_CMAKE_IN_TRY_COMPILE" not in content

        # remove multiple
        toolchain = CMakeToolchain(conanfile)
        toolchain.blocks.select("generic_system", "cmake_flags_init")
        content = toolchain.content
        assert "########## 'generic_system' block #############" in content
        assert "########## 'cmake_flags_init' block #############" in content
        assert "########## 'libcxx' block #############" not in content
        # These are not removed by default, to not break behavior
        assert "########## 'variables' block #############" in content
        assert "########## 'preprocessor' block #############" in content
        assert 'CMAKE_SYSTEM_NAME' in content
        assert "CMAKE_CXX_FLAGS_INIT" in content
        assert "_CMAKE_IN_TRY_COMPILE" not in content

        # remove multiple
        toolchain = CMakeToolchain(conanfile)
        toolchain.blocks.enabled("generic_system")
        content = toolchain.content
        assert "########## 'generic_system' block #############" in content
        assert "########## 'variables' block #############" not in content
        assert "########## 'preprocessor' block #############" not in content

    def test_enabled_blocks_conf(self, conanfile):
        conanfile.conf.define("tools.cmake.cmaketoolchain:enabled_blocks", ["generic_system"])
        toolchain = CMakeToolchain(conanfile)
        content = toolchain.content
        assert "########## 'generic_system' block #############" in content
        assert "########## 'cmake_flags_init' block #############" not in content
        assert "########## 'libcxx' block #############" not in content
        assert "########## 'variables' block #############" not in content
        assert "########## 'preprocessor' block #############" not in content

        # remove multiple
        conanfile.conf.define("tools.cmake.cmaketoolchain:enabled_blocks",
                              ["generic_system", "cmake_flags_init"])
        toolchain = CMakeToolchain(conanfile)
        content = toolchain.content
        assert "########## 'generic_system' block #############" in content
        assert "########## 'cmake_flags_init' block #############" in content
        assert "########## 'libcxx' block #############" not in content
        assert "########## 'variables' block #############" not in content
        assert "########## 'preprocessor' block #############" not in content

        conanfile.conf.define("tools.cmake.cmaketoolchain:enabled_blocks", ["potato"])
        toolchain = CMakeToolchain(conanfile)
        with pytest.raises(ConanException) as e:
            _ = toolchain.content
        assert "Block 'potato' defined in tools.cmake.cmaketoolchain:enabled_blocks" in str(e)

    def test_dict_keys(self, conanfile):
        toolchain = CMakeToolchain(conanfile)
        assert "generic_system" in toolchain.blocks.keys()
        items = dict(toolchain.blocks.items())
        assert "generic_system" in items

    def test_template_remove(self, conanfile):
        toolchain = CMakeToolchain(conanfile)
        toolchain.blocks["generic_system"].template = ""
        content = toolchain.content
        assert 'CMAKE_SYSTEM_NAME' not in content

    def test_template_change(self, conanfile):
        toolchain = CMakeToolchain(conanfile)
        tmp = toolchain.blocks["generic_system"].template
        toolchain.blocks["generic_system"].template = tmp.replace("CMAKE_SYSTEM_NAME", "OTHER_THING")
        content = toolchain.content
        assert 'set(OTHER_THING potato)' in content

    def test_context_change(self, conanfile):
        toolchain = CMakeToolchain(conanfile)
        tmp = toolchain.blocks["generic_system"]

        def context(self_):
            assert self_
            return {"cmake_system_name": None}

        tmp.context = types.MethodType(context, tmp)
        content = toolchain.content
        assert 'CMAKE_SYSTEM_NAME' not in content

    def test_context_update(self, conanfile):
        toolchain = CMakeToolchain(conanfile)
        cmake_system_name = toolchain.blocks["generic_system"].values["cmake_system_name"]
        toolchain.blocks["generic_system"].values["cmake_system_name"] = "Super" + cmake_system_name
        content = toolchain.content
        assert 'set(CMAKE_SYSTEM_NAME Superpotato)' in content

    def test_context_replace(self, conanfile):
        toolchain = CMakeToolchain(conanfile)
        toolchain.blocks["generic_system"].values = {"cmake_system_name": "SuperPotato"}
        content = toolchain.content
        assert 'set(CMAKE_SYSTEM_NAME SuperPotato)' in content

    def test_replace_block(self, conanfile):
        toolchain = CMakeToolchain(conanfile)

        class MyBlock(Block):
            template = "HelloWorld"

            def context(self):
                return {}

        toolchain.blocks["generic_system"] = MyBlock
        content = toolchain.content
        assert 'HelloWorld' in content
        assert 'set(CMAKE_SYSTEM_NAME potato)' not in content

    def test_add_new_block(self, conanfile):
        toolchain = CMakeToolchain(conanfile)

        class MyBlock(Block):
            template = "Hello {{myvar}}!!!"

            def context(self):
                return {"myvar": "World"}

        toolchain.blocks["mynewblock"] = MyBlock
        content = toolchain.content
        assert 'Hello World!!!' in content
        assert 'set(CMAKE_SYSTEM_NAME potato)' in content

    def test_user_toolchain(self, conanfile):
        toolchain = CMakeToolchain(conanfile)
        toolchain.blocks["user_toolchain"].values["paths"] = ["myowntoolchain.cmake"]
        content = toolchain.content
        assert 'include("myowntoolchain.cmake")' in content


class TestBlocksFunctionality:
    def test_cmake_toolchain(self, conanfile):
        toolchain = CMakeToolchain(conanfile)
        content = toolchain.content
        assert 'set(CMAKE_SYSTEM_NAME potato)' in content

    def test_compilers_block(self, conanfile):
        cmake_mapping = {"c": "C", "cuda": "CUDA", "cpp": "CXX", "objc": "OBJC",
                         "objcpp": "OBJCXX", "rc": "RC", 'fortran': "Fortran", 'asm': "ASM",
                         "hip": "HIP", "ispc": "ISPC"}
        compilers = {"c": "path_to_c", "cuda": "path_to_cuda", "cpp": "path_to_cpp",
                     "objc": "path_to_objc", "objcpp": "path_to_objcpp", "rc": "path_to_rc",
                     'fortran': "path_to_fortran", 'asm': "path_to_asm", "hip": "path_to_hip",
                     "ispc": "path_to_ispc"}
        conanfile.conf.define("tools.build:compiler_executables", compilers)
        toolchain = CMakeToolchain(conanfile)
        content = toolchain.content
        for compiler, lang in cmake_mapping.items():
            assert f'set(CMAKE_{lang}_COMPILER "path_to_{compiler}")' in content

    def test_linker_scripts_block(self, conanfile):
        conanfile.conf.define("tools.build:linker_scripts",
                              ["path_to_first_linker_script", "path_to_second_linker_script"])
        toolchain = CMakeToolchain(conanfile)
        content = toolchain.content
        assert r'string(APPEND CONAN_EXE_LINKER_FLAGS " -T\"path_to_first_linker_script\" ' \
               r'-T\"path_to_second_linker_script\"")' in content

    def test_libcxx_abi_flag(self, base_conanfile):
        c = base_conanfile
        c.settings.build_type = "Release"
        c.settings.arch = "x86_64"
        c.settings.compiler = "gcc"
        c.settings.compiler.version = "11"
        c.settings.compiler.cppstd = "20"
        c.settings.compiler.libcxx = "libstdc++"
        c.settings.os = "Linux"

        toolchain = CMakeToolchain(c)
        content = toolchain.content
        assert '_GLIBCXX_USE_CXX11_ABI=0' in content
        c.settings.compiler.libcxx = "libstdc++11"
        toolchain = CMakeToolchain(c)
        content = toolchain.content
        # by default, no flag is output anymore, it is assumed the compiler default
        assert 'GLIBCXX_USE_CXX11_ABI' not in content
        # recipe workaround for older distros
        toolchain.blocks["libcxx"].values["glibcxx"] = "_GLIBCXX_USE_CXX11_ABI=1"
        content = toolchain.content
        assert '_GLIBCXX_USE_CXX11_ABI=1' in content

        # but maybe the conf is better
        c.conf.define("tools.gnu:define_libcxx11_abi", True)
        toolchain = CMakeToolchain(c)
        content = toolchain.content
        assert '_GLIBCXX_USE_CXX11_ABI=1' in content


class TestMSVCToolset:
    @pytest.fixture
    def conanfile_msvc(self, base_conanfile):
        c = base_conanfile
        c.settings.build_type = "Release"
        c.settings.arch = "x86"
        c.settings.compiler = "msvc"
        c.settings.compiler.version = "194"
        c.settings.compiler.cppstd = "20"
        c.settings.os = "Windows"
        return c

    def test_toolset(self, conanfile_msvc):
        toolchain = CMakeToolchain(conanfile_msvc)
        assert 'set(CMAKE_GENERATOR_TOOLSET "v143" CACHE STRING "" FORCE)' in toolchain.content
        assert 'Visual Studio 17 2022' in toolchain.generator
        assert 'CMAKE_CXX_STANDARD 20' in toolchain.content

    def test_toolset_latest_generator(self, conanfile_msvc):
        conanfile_msvc.settings.compiler.version = "195"
        toolchain = CMakeToolchain(conanfile_msvc)
        assert 'set(CMAKE_GENERATOR_TOOLSET "v145" CACHE STRING "" FORCE)' in toolchain.content
        assert 'Visual Studio 18 2026' in toolchain.generator

    def test_toolset_update_version(self, conanfile_msvc):
        conanfile_msvc.settings.compiler.version = "193"
        conanfile_msvc.settings.compiler.update = "8"
        c = CMakeToolchain(conanfile_msvc)
        assert 'set(CMAKE_GENERATOR_TOOLSET "v143,version=14.38" CACHE STRING "" FORCE)' in c.content

    def test_toolset_update_version_conf(self, conanfile_msvc):
        conanfile_msvc.settings.compiler.version = "193"
        conanfile_msvc.conf.define("tools.microsoft:msvc_update", "7")
        c = CMakeToolchain(conanfile_msvc)
        assert 'set(CMAKE_GENERATOR_TOOLSET "v143,version=14.37" CACHE STRING "" FORCE)' in c.content

    def test_toolset_update_version_forced_conf(self, conanfile_msvc):
        conanfile_msvc.settings.compiler.version = "193"
        conanfile_msvc.settings.compiler.update = "8"
        conanfile_msvc.conf.define("tools.microsoft:msvc_update", "7")
        c = CMakeToolchain(conanfile_msvc)
        assert 'set(CMAKE_GENERATOR_TOOLSET "v143,version=14.37" CACHE STRING "" FORCE)' in c.content

    def test_toolset_update_version_overflow(self, conanfile_msvc):
        # https://github.com/conan-io/conan/issues/15583
        conanfile_msvc.settings.compiler.version = "194"
        conanfile_msvc.settings.compiler.update = "8"
        c = CMakeToolchain(conanfile_msvc)
        assert 'set(CMAKE_GENERATOR_TOOLSET "v143,version=14.48" CACHE STRING "" FORCE)' in c.content

    def test_toolset_x64(self, conanfile_msvc):
        # https://github.com/conan-io/conan/issues/11144
        conanfile_msvc.conf.define("tools.cmake.cmaketoolchain:toolset_arch", "x64")
        c = CMakeToolchain(conanfile_msvc)
        assert 'set(CMAKE_GENERATOR_TOOLSET "v143,host=x64" CACHE STRING "" FORCE)' in c.content
        assert 'Visual Studio 17 2022' in c.generator
        assert 'CMAKE_CXX_STANDARD 20' in c.content

    def test_toolset_cuda(self, conanfile_msvc):
        conanfile_msvc.conf.define("tools.cmake.cmaketoolchain:toolset_cuda", "C:/Path/To/CUDA")
        c = CMakeToolchain(conanfile_msvc)
        assert ('set(CMAKE_GENERATOR_TOOLSET "v143,cuda=C:/Path/To/CUDA" '
                'CACHE STRING "" FORCE)') in c.content

    def test_older_msvc_toolset(self, conanfile_msvc):
        c = conanfile_msvc
        c.settings.compiler.version = "170"
        c.settings.compiler.cppstd = "14"
        toolchain = CMakeToolchain(c)
        content = toolchain.content
        assert 'CMAKE_GENERATOR_TOOLSET "v110"' in content
        # As by the CMake docs, this has no effect for VS < 2015
        assert 'CMAKE_CXX_STANDARD 14' in content

    def test_vs_clangcl(self, conanfile_msvc):
        c = conanfile_msvc
        c.settings.build_type = "Release"
        c.settings.arch = "x86_64"
        c.settings.compiler = "clang"
        c.settings.compiler.runtime_version = "v145"
        c.settings.compiler.version = "20"
        c.settings.compiler.cppstd = "14"
        c.settings.os = "Windows"
        c.conf.define("tools.cmake.cmaketoolchain:generator", "Visual Studio 18")
        toolchain = CMakeToolchain(c)
        content = toolchain.content
        assert 'CMAKE_GENERATOR_TOOLSET "ClangCL"' in content
        assert 'CMAKE_CXX_STANDARD 14' in content

    def test_older_msvc_toolset_update(self, conanfile_msvc):
        # https://github.com/conan-io/conan/issues/15787
        c = conanfile_msvc
        c.settings.arch = "x86_64"
        c.settings.compiler = "msvc"
        c.settings.compiler.version = "192"
        c.settings.compiler.update = "9"
        c.settings.compiler.cppstd = "14"
        c.settings.os = "Windows"
        toolchain = CMakeToolchain(c)
        content = toolchain.content
        assert 'CMAKE_GENERATOR_TOOLSET "v142,version=14.29"' in content

    def test_msvc_xp_toolsets(self, conanfile_msvc):
        c = conanfile_msvc
        c.settings.build_type = "Release"
        c.settings.arch = "x86"
        c.settings.compiler = "msvc"
        c.settings.compiler.version = "170"
        c.settings.compiler.toolset = "v110_xp"
        c.settings.compiler.cppstd = "14"
        toolchain = CMakeToolchain(c)
        content = toolchain.content
        assert 'CMAKE_GENERATOR_TOOLSET "v110_xp"' in content
        # As by the CMake docs, this has no effect for VS < 2015
        assert 'CMAKE_CXX_STANDARD 14' in content


class TestFpic:
    @pytest.fixture
    def conanfile_linux(self, base_conanfile):
        c = base_conanfile
        c.settings.build_type = "Release"
        c.settings.arch = "x86_64"
        c.settings.compiler = "gcc"
        c.settings.compiler.version = "11"
        c.settings.compiler.cppstd = "20"
        c.settings.os = "Linux"
        return c

    def test_no_fpic_when_not_an_option(self, conanfile_linux):
        toolchain = CMakeToolchain(conanfile_linux)
        content = toolchain.content
        assert 'set(CMAKE_POSITION_INDEPENDENT_CODE' not in content

    @pytest.fixture
    def conanfile_linux_shared(self, conanfile_linux):
        c = conanfile_linux
        c.options = Options({"fPIC": [True, False],
                             "shared": [True, False]},
                            {"fPIC": False, "shared": True, })
        return c

    @pytest.mark.parametrize("fpic", [True, False])
    def test_fpic_when_shared_true(self, conanfile_linux_shared, fpic):
        conanfile_linux_shared.options.fPIC = fpic
        toolchain = CMakeToolchain(conanfile_linux_shared)
        cmake_value = 'ON' if fpic else 'OFF'
        content = toolchain.content
        assert 'set(CMAKE_POSITION_INDEPENDENT_CODE {} CACHE BOOL'.format(cmake_value) in content

    def test_fpic_when_not_shared(self, conanfile_linux_shared):
        conanfile_linux_shared.options.shared = False
        toolchain = CMakeToolchain(conanfile_linux_shared)
        content = toolchain.content
        assert 'set(CMAKE_POSITION_INDEPENDENT_CODE' in content

    @pytest.fixture
    def conanfile_windows_fpic(self, base_conanfile):
        c = base_conanfile
        c.options = Options({"fPIC": [True, False]},
                            {"fPIC": True})
        c.settings.build_type = "Release"
        c.settings.arch = "x86"
        c.settings.compiler = "gcc"
        c.settings.compiler.libcxx = "libstdc++"
        c.settings.os = "Windows"
        return c

    def test_no_fpic_on_windows(self, conanfile_windows_fpic):
        toolchain = CMakeToolchain(conanfile_windows_fpic)
        content = toolchain.content
        assert 'set(CMAKE_POSITION_INDEPENDENT_CODE' not in content

    @pytest.fixture
    def conanfile_linux_fpic(self, base_conanfile):
        c = base_conanfile
        c.options = Options({"fPIC": [True, False]},
                            {"fPIC": False})
        c.settings.build_type = "Release"
        c.settings.arch = "x86_64"
        c.settings.compiler = "gcc"
        c.settings.compiler.version = "11"
        c.settings.compiler.cppstd = "20"
        c.settings.os = "Linux"
        return c

    def test_fpic_disabled(self, conanfile_linux_fpic):
        conanfile_linux_fpic.options.fPIC = False
        toolchain = CMakeToolchain(conanfile_linux_fpic)
        content = toolchain.content
        assert 'set(CMAKE_POSITION_INDEPENDENT_CODE OFF' in content

    def test_fpic_enabled(self, conanfile_linux_fpic):
        conanfile_linux_fpic.options.fPIC = True
        toolchain = CMakeToolchain(conanfile_linux_fpic)
        content = toolchain.content
        assert 'set(CMAKE_POSITION_INDEPENDENT_CODE ON' in content


class TestApple:
    def test_osx_deployment_target(self, base_conanfile):
        c = base_conanfile
        c.settings.build_type = "Release"
        c.settings.arch = "x86"
        c.settings.compiler = "apple-clang"
        c.settings.compiler.libcxx = "libc++"
        c.settings.os = "Macos"
        c.settings.os.version = "10.15"
        toolchain = CMakeToolchain(c)
        content = toolchain.content
        assert 'set(CMAKE_OSX_DEPLOYMENT_TARGET "10.15" CACHE STRING "")' in content

    @pytest.mark.parametrize("os,os_sdk,arch,expected_sdk", [
        ("Macos", None, "x86_64", "macosx"),
        ("Macos", None, "armv7", "macosx"),
        ("iOS", "iphonesimulator", "armv8", "iphonesimulator"),
        ("watchOS", "watchsimulator", "armv8", "watchsimulator")
    ])
    def test_apple_cmake_osx_sysroot(self, base_conanfile, os, os_sdk, arch, expected_sdk):
        """
        Testing if CMAKE_OSX_SYSROOT is correctly set.
        Issue related: https://github.com/conan-io/conan/issues/10275
        """
        c = base_conanfile
        c.settings.os = os
        if os_sdk:
            c.settings.os.sdk = os_sdk
        c.settings.build_type = "Release"
        c.settings.arch = arch
        c.settings.compiler = "apple-clang"
        c.settings.compiler.version = "13.0"
        c.settings.compiler.libcxx = "libc++"
        c.settings.compiler.cppstd = "17"

        toolchain = CMakeToolchain(c)
        content = toolchain.content
        assert 'set(CMAKE_OSX_SYSROOT %s CACHE STRING "" FORCE)' % expected_sdk in content

    @pytest.mark.parametrize("os,arch,expected_sdk", [
        ("iOS", "x86_64", ""),
        ("watchOS", "armv8", ""),
        ("tvOS", "x86_64", "")
    ])
    def test_apple_cmake_osx_sysroot_sdk_mandatory(self, base_conanfile, os, arch, expected_sdk):
        """
        Testing if CMAKE_OSX_SYSROOT is correctly set.
        Issue related: https://github.com/conan-io/conan/issues/10275
        """
        c = base_conanfile
        c.settings.os = os
        c.settings.build_type = "Release"
        c.settings.arch = arch
        c.settings.compiler = "apple-clang"
        c.settings.compiler.version = "13.0"
        c.settings.compiler.libcxx = "libc++"
        c.settings.compiler.cppstd = "17"

        with pytest.raises(ConanException) as excinfo:
            _ = CMakeToolchain(c).content
        assert "Please, specify a suitable value for os.sdk." in str(excinfo.value)


class TestCrossBuild:
    @pytest.fixture
    def conanfile_cross(self, base_conanfile):
        c = base_conanfile
        c.settings.build_type = "Release"
        c.settings.arch = "armv8"
        c.settings.compiler = "gcc"
        c.settings.compiler.version = "11"
        c.settings.compiler.cppstd = "20"
        c.settings.os = "baremetal"
        c.settings_build = c.settings.copy()
        c.settings_build.os = "Linux"
        c.settings_build.arch = "x86_64"
        return c

    def test_cmake_system_name(self, conanfile_cross):
        toolchain = CMakeToolchain(conanfile_cross)
        content = toolchain.content
        assert 'set(CMAKE_SYSTEM_NAME Generic)' in content



import json
import os
import textwrap

from conan.cli.exit_codes import ERROR_GENERAL
from conan.test.utils.tools import NO_SETTINGS_PACKAGE_ID, TestClient, GenConanfile


class TestBasicCliOutput:

    def test_info_settings(self):
        client = TestClient()
        conanfile = textwrap.dedent("""
            from conan import ConanFile

            class MyTest(ConanFile):
                name = "pkg"
                version = "0.2"
                settings = "build_type", "compiler"
                author = "John Doe"
                license = "MIT"
                url = "https://foo.bar.baz"
                homepage = "https://foo.bar.site"
                topics = "foo", "bar", "qux"
                provides = "libjpeg", "libjpg"
                deprecated = "other-pkg"
                options = {"shared": [True, False], "fPIC": [True, False]}
                default_options = {"shared": False, "fPIC": True}
            """)
        client.save({"conanfile.py": conanfile})
        client.run("graph info . -s build_type=Debug -s compiler=gcc -s compiler.version=11")
        assert "build_type: Debug" in client.out
        assert "context: host" in client.out
        assert "license: MIT" in client.out
        assert "homepage: https://foo.bar.site" in client.out
        assert "compiler: gcc" in client.out
        assert "compiler.version: 11" in client.out
        assert "prev: None" in client.out
        assert "fPIC: True" in client.out
        assert "shared: False" in client.out
        # Create the package
        client.run("create .")
        pref = client.get_latest_package_reference("pkg/0.2")
        client.run("graph info .")
        # It's a consumer so we can not get the prev because it's not even a package yet
        assert "prev: None" in client.out
        # Now, let's create another consumer requiring the previous package
        client.save({"conanfile.txt": "[requires]\npkg/0.2"}, clean_first=True)
        client.run("graph info .")
        assert textwrap.dedent(f"""
            {repr(pref.ref)}:
              ref: {repr(pref.ref)}
              id: 1
              recipe: Cache
              package_id: {pref.package_id}
              prev: {pref.revision}""") in client.out

    def test_nontuple_topics(self):
        client = TestClient()
        # This is the representation that should always happen,
        # we wouldn't expect topics not to be a tuple here
        conanfile = textwrap.dedent("""
                            from conan import ConanFile

                            class MyTest(ConanFile):
                                name = "pkg"
                                version = "0.2"
                                provides = ("bar",)
                                topics = ("foo",)
                            """)
        client.save({"conanfile.py": conanfile})
        client.run("graph info . --format=json")
        recipe = json.loads(client.stdout)["graph"]["nodes"]["0"]
        assert type(recipe["topics"]) == list
        assert recipe["topics"] == ["foo"]
        assert type(recipe["provides"]) == list
        assert recipe["provides"] == ["bar"]

        # But this used to fail,
        # topics were not converted to a list internally if one was not provided
        client2 = TestClient()
        conanfile2 = textwrap.dedent("""
                    from conan import ConanFile

                    class MyTest(ConanFile):
                        name = "pkg"
                        version = "0.2"
                        provides = "bar"
                        topics = "foo"
                    """)
        client2.save({"conanfile.py": conanfile2})
        client2.run("graph info . --format=json")
        recipe = json.loads(client2.stdout)["graph"]["nodes"]["0"]
        assert type(recipe["topics"]) == list
        assert recipe["topics"] == ["foo"]
        assert type(recipe["provides"]) == list
        assert recipe["provides"] == ["bar"]


class TestConanfilePath:
    def test_cwd(self):
        # Check the first positional argument is a relative path
        client = TestClient()
        conanfile = GenConanfile("pkg", "0.1").with_setting("build_type")
        client.save({"subfolder/conanfile.py": conanfile})
        client.run("graph info ./subfolder -s build_type=Debug")
        assert "build_type: Debug" in client.out

    def test_wrong_path_parameter(self):
        # check that the positional parameter must exist and file must be found
        client = TestClient()

        client.run("graph info", assert_error=True)
        assert "ERROR: Conanfile not found" in client.out

        client.run("graph info not_real_path", assert_error=True)
        assert "ERROR: Conanfile not found" in client.out

        client.run("graph info conanfile.txt", assert_error=True)
        assert "ERROR: Conanfile not found" in client.out

        client.run("graph info --requires=foo/1.0", assert_error=True)
        assert "ERROR: Package 'foo/1.0' not resolved: No remote defined." in client.out

        client.run("graph info . --requires=foo/1.0", assert_error=True)
        assert "--requires and --tool-requires arguments are incompatible with [path] '.' argument" in client.out


class TestFilters:

    def test_filter_fields(self):
        # The --filter arg should work, specifying which fields to show only
        c = TestClient()
        c.save({"conanfile.py": GenConanfile()
               .with_class_attribute("author = 'myself'")
               .with_class_attribute("license = 'MIT'")
               .with_class_attribute("url = 'https://url.com'")})
        c.run("graph info . ")
        assert "license: MIT" in c.out
        assert "author: myself" in c.out
        c.run("graph info . --filter=license")
        assert "license: MIT" in c.out
        assert "author" not in c.out
        c.run("graph info . --filter=author")
        assert "license" not in c.out
        assert "author: myself" in c.out
        c.run("graph info . --filter=author --filter=license")
        assert "license" in c.out
        assert "author: myself" in c.out

    def test_filter_fields_json(self):
        # The --filter arg should work, specifying which fields to show only
        c = TestClient()
        c.save({"conanfile.py": GenConanfile()
               .with_class_attribute("author = 'myself'")
               .with_class_attribute("license = 'MIT'")
               .with_class_attribute("url = 'https://url.com'")})
        c.run("graph info . --filter=license --format=json")
        assert "author" not in c.out
        assert '"license": "MIT"' in c.out
        c.run("graph info . --filter=license --format=html", assert_error=True)
        assert "Formatted output 'html' cannot filter fields" in c.out


class TestJsonOutput:
    def test_json_package_filter(self):
        client = TestClient()
        conanfile = GenConanfile("pkg", "0.1").with_setting("build_type").with_requires("dep/0.1")
        client.save({"dep/conanfile.py": GenConanfile("dep", "0.1"),
                     "conanfile.py": conanfile})
        client.run("create dep")
        client.run("graph info . --package-filter=nothing --format=json")
        assert '"nodes": {}' in client.out
        client.run("graph info . --package-filter=pkg* --format=json")
        graph = json.loads(client.stdout)
        assert len(graph["graph"]["nodes"]) == 1
        assert graph["graph"]["nodes"]["0"]["ref"] == "pkg/0.1"

        # Consumer & filtering
        client.run("graph info . --package-filter=& --format=json")
        graph = json.loads(client.stdout)
        assert len(graph["graph"]["nodes"]) == 1
        assert graph["graph"]["nodes"]["0"]["ref"] == "pkg/0.1"

        # This returns nothing, doesn't work that way, serialized graph doesn't have the consumer
        client.run("graph info --requires=dep/0.1 --package-filter=& --format=json")
        graph = json.loads(client.stdout)
        assert len(graph["graph"]["nodes"]) == 1
        assert graph["graph"]["nodes"]["1"]["name"] == "dep"

        # Filter matches all
        client.run("graph info . --package-filter=* --format=json")
        graph = json.loads(client.stdout)
        assert len(graph["graph"]["nodes"]) == 2
        assert graph["graph"]["nodes"]["0"]["ref"] == "pkg/0.1"
        assert graph["graph"]["nodes"]["1"]["name"] == "dep"

    def test_json_info_outputs(self):
        client = TestClient()
        conanfile = GenConanfile("pkg", "0.1").with_setting("build_type")
        client.save({"conanfile.py": conanfile})
        client.run("graph info . -s build_type=Debug --format=json")
        graph = json.loads(client.stdout)
        assert graph["graph"]["nodes"]["0"]["settings"]["build_type"] == "Debug"

    def test_json_info_requirements(self):
        c = TestClient()
        c.save({"pkg/conanfile.py": GenConanfile("pkg", "0.1"),
                "app/conanfile.py": GenConanfile().with_require("pkg/[>=0.0 <1.0]")})
        c.run("create pkg")
        c.run("graph info app --format=json")
        graph = json.loads(c.stdout)
        assert graph["graph"]["nodes"]["0"]["dependencies"]["1"]["require"] == "pkg/[>=0.0 <1.0]"


class TestAdvancedCliOutput:
    """ Testing more advanced fields output, like SCM or PYTHON-REQUIRES
    """

    def test_python_requires(self):
        # https://github.com/conan-io/conan/issues/9277
        client = TestClient()
        client.save({"conanfile.py": GenConanfile()})
        client.run("export . --name=tool --version=0.1")
        conanfile = textwrap.dedent("""
            from conan import ConanFile

            class pkg(ConanFile):
                python_requires = "tool/0.1"
            """)
        client.save({"conanfile.py": conanfile})

        client.run("graph info .")
        assert "python_requires:\n    tool/0.1#4d670581ccb765839f2239cc8dff8fbd" in client.out

        client.run("graph info . --format=json")
        info = json.loads(client.stdout)
        pyrequires = info["graph"]["nodes"]["0"]["python_requires"]
        tool = pyrequires["tool/0.1#4d670581ccb765839f2239cc8dff8fbd"]
        info = info["graph"]["nodes"]["0"]["info"]
        assert info["python_requires"] == ["tool/0.1.Z"]
        # lets make sure the path exists
        assert tool["recipe"] == "Cache"
        assert tool["remote"] is None
        assert os.path.exists(tool["path"])

    def test_build_id_info(self):
        client = TestClient()
        conanfile = textwrap.dedent("""
            from conan import ConanFile

            class Pkg(ConanFile):
                name = "pkg"
                version = "0.1"
                settings = "build_type"

                def build_id(self):
                    self.info_build.settings.build_type = "Any"
            """)
        client.save({"conanfile.py": conanfile})
        client.run("export .")
        client.save({"conanfile.py": GenConanfile().with_requires("pkg/0.1")}, clean_first=True)
        client.run("graph info . -s build_type=Release")
        assert "build_id: ec0cd314abe055f7de86cd6493e31977d2b87884" in client.out
        assert "package_id: efa83b160a55b033c4ea706ddb980cd708e3ba1b" in client.out

        client.run("graph info . -s build_type=Debug")
        assert "build_id: ec0cd314abe055f7de86cd6493e31977d2b87884" in client.out
        assert "package_id: 9e186f6d94c008b544af1569d1a6368d8339efc5" in client.out


class TestEditables:
    def test_info_paths(self):
        # https://github.com/conan-io/conan/issues/7054
        c = TestClient()
        conanfile = textwrap.dedent("""
            from conan import ConanFile
            class Pkg(ConanFile):
                def layout(self):
                    self.folders.source = "."
                    self.folders.build = "."
            """)
        c.save({"pkg/conanfile.py": conanfile,
                "consumer/conanfile.py": GenConanfile().with_require("pkg/0.1")})
        c.run("editable add pkg --name=pkg --version=0.1")
        # TODO: Check this --package-filter with *
        c.run("graph info consumer --package-filter=pkg*")
        # FIXME: Paths are not diplayed yet
        assert "source_folder: None" in c.out


class TestInfoTestPackage:
    # https://github.com/conan-io/conan/issues/10714

    def test_tested_reference_str(self):
        client = TestClient()
        client.save({"conanfile.py": GenConanfile("tool", "0.1")})
        client.run("export .")

        conanfile = textwrap.dedent("""
            from conan import ConanFile
            class HelloConan(ConanFile):

                def requirements(self):
                    self.requires(self.tested_reference_str)

                def build_requirements(self):
                    self.build_requires(self.tested_reference_str)
            """)
        client.save({"conanfile.py": conanfile})

        for args in ["", " --build=*"]:
            client.run("graph info . " + args)
            assert "AttributeError: 'HelloConan' object has no attribute 'tested_reference_str'"\
                   not in client.out


class TestDeployers:

    def test_custom_deploy(self):
        c = TestClient()
        conanfile = GenConanfile("pkg", "0.1").with_class_attribute("license = 'MIT'")
        c.save({"conanfile.py": conanfile})
        c.run("create .")
        collectlicenses = textwrap.dedent(r"""
            from conan.tools.files import save

            def deploy(graph, output_folder, **kwargs):
                contents = []
                conanfile = graph.root.conanfile
                for pkg in graph.nodes:
                    d = pkg.conanfile
                    contents.append("LICENSE {}: {}!".format(d.ref, d.license))
                contents = "\n".join(contents)
                conanfile.output.info(contents)
                save(conanfile, "licenses.txt", contents)
            """)
        c.save({"conanfile.py": GenConanfile().with_requires("pkg/0.1")
                                              .with_class_attribute("license='GPL'"),
                "collectlicenses.py": collectlicenses})
        c.run("graph info . --deployer=collectlicenses")
        assert "conanfile.py: LICENSE : GPL!" in c.out
        assert "LICENSE pkg/0.1: MIT!" in c.out
        contents = c.load("licenses.txt")
        assert "LICENSE pkg/0.1: MIT!" in contents
        assert "LICENSE : GPL!" in contents


class TestErrorsInGraph:
    # https://github.com/conan-io/conan/issues/12735

    def test_error_in_recipe(self):
        c = TestClient()
        dep = textwrap.dedent("""
            from conan import ConanFile
            class Pkg(ConanFile):
                name = "dep"
                version = "0.1"
                def package_id(self):
                    a = b
            """)
        c.save({"dep/conanfile.py": dep,
                "consumer/conanfile.py": GenConanfile().with_requires("dep/0.1")})
        c.run("export dep")
        exit_code = c.run("graph info consumer", assert_error=True)
        assert "name 'b' is not defined" in c.out
        assert exit_code == ERROR_GENERAL

    def test_error_exports(self):
        c = TestClient()
        dep = textwrap.dedent("""
            from conan import ConanFile
            from conan.tools.files import replace_in_file
            class Pkg(ConanFile):
                name = "dep"
                version = "0.1"
                def export(self):
                    replace_in_file(self, "conanfile.py", "from conan", "from conans")
            """)
        c.save({"dep/conanfile.py": dep,
                "consumer/conanfile.py": GenConanfile().with_requires("dep/0.1")})
        c.run("export dep")
        exit_code = c.run("graph info consumer", assert_error=True)
        assert "ERROR: Package 'dep/0.1' not resolved: dep/0.1: Cannot load" in c.out
        assert exit_code == ERROR_GENERAL

    def test_missing_invalid(self):
        c = TestClient()
        dep_invalid = textwrap.dedent(
            """
           from conan import ConanFile
           from conan.errors import ConanInvalidConfiguration
           class Pkg(ConanFile):
               name = "dep_invalid"
               settings = "os", "build_type"
               version = "0.1"
               def validate(self):
                   raise ConanInvalidConfiguration("Invalid package")
        """
        )

        c.save({
                "dep/conanfile.py": GenConanfile("dep").with_version("0.1"),
                "dep_invalid/conanfile.py": dep_invalid,
                "consumer/conanfile.py": GenConanfile("consumer")
                                        .with_requires("dep/0.1")
                                        .with_requires("dep_invalid/0.1"),
            })
        c.run("export dep")
        c.run("export dep_invalid")
        exit_code = c.run("graph info consumer")
        assert "WARN: There are some error(s) in the graph:" in c.out
        assert "- Missing packages: dep" in c.out
        assert "- Invalid packages: dep_invalid" in c.out
        assert exit_code == 0


class TestInfoUpdate:

    def test_update(self):
        c = TestClient(default_server_user=True)
        c.save({"conanfile.py": GenConanfile("tool")})
        c.run("create . --version=1.0")
        c.run("create . --version=1.1")
        c.run("upload tool/1.1 -r=default -c")
        c.run("remove tool/1.1 -c")
        c.save({"conanfile.py": GenConanfile().with_requires("tool/[*]")})
        c.run("graph info . --filter=recipe")
        assert "tool/1.0#7fbd52996f34447f4a4c362edb5b4af5 - Cache" in c.out
        c.run("graph info . --update --filter=recipe")
        assert "tool/1.1#7fbd52996f34447f4a4c362edb5b4af5 - Downloaded (default)" in c.out


def test_info_not_hit_server():
    """
    the command graph info shouldn't be hitting the server if packages are in the Conan cache
    :return:
    """
    c = TestClient(default_server_user=True)
    c.save({"pkg/conanfile.py": GenConanfile("pkg", "0.1"),
            "consumer/conanfile.py": GenConanfile("consumer", "0.1").with_require("pkg/0.1")})
    c.run("create pkg")
    c.run("create consumer")
    c.run("upload * -r=default -c")
    c.run("remove * -c")
    c.run("install --requires=consumer/0.1@")
    assert "Downloaded" in c.out
    # break the server to make sure it is not being contacted at all
    c.servers["default"] = None
    c.run("graph info --requires=consumer/0.1@")
    assert "Downloaded" not in c.out
    # Now we remove the local, so it will raise errors
    c.run("remove pkg* -c")
    c.run("graph info --requires=consumer/0.1@", assert_error=True)
    assert "'NoneType' object " in c.out
    c.run("remote disable *")
    c.run("graph info --requires=consumer/0.1@", assert_error=True)
    assert "'NoneType' object " not in c.out
    assert "No remote defined" in c.out


def test_graph_info_user():
    """
    https://github.com/conan-io/conan/issues/15791
    """
    c = TestClient()
    conanfile = textwrap.dedent("""
        from conan import ConanFile
        class Pkg(ConanFile):
            name = "pkg"
            version = "0.1"
            user = "user"
        """)

    c.save({"conanfile.py": conanfile})
    c.run("graph info .")
    assert "pkg/0.1@user" in c.out
    c.run("graph info . --channel=channel")
    assert "pkg/0.1@user/channel" in c.out


def test_graph_info_bundle():
    c = TestClient(light=True)
    c.save({"subfolder/conanfile.py": GenConanfile("liba", "1.0")})
    c.run("create ./subfolder")
    conanfile = textwrap.dedent("""
        from conan import ConanFile
        class RepackageRecipe(ConanFile):
            name = "lib"
            version = "1.0"
            def requirements(self):
                self.requires("liba/1.0")
            vendor = True
        """)
    c.save({"conanfile.py": conanfile})
    c.run("create .")
    c.save({"conanfile.py": GenConanfile("consumer", "1.0").with_requires("lib/1.0")})

    c.run("graph info . --build='lib*'")
    c.assert_listed_binary({"lib/1.0": (NO_SETTINGS_PACKAGE_ID, "Invalid")})

    c.run("graph info . -c tools.graph:vendor=build --build='lib*'")
    c.assert_listed_binary({"lib/1.0": (NO_SETTINGS_PACKAGE_ID, "Build")})


def test_graph_info_provides_conflict_html():
    """ This test also makes sure that the serialization of the GraphProvidesConflict works"""
    tc = TestClient(light=True)
    tc.save({"bar/conanfile.py": GenConanfile("bar", "1.0"),
             "foo/conanfile.py": GenConanfile("foo", "1.0").with_provides("bar")})

    tc.run("export bar")
    tc.run("export foo")
    tc.run("graph info --requires=bar/1.0 --requires=foo/1.0 --format=html", assert_error=True)
    # It got properly serialized
    assert '"type": "provide_conflict"' in tc.out
    assert "Both 'bar/1.0' and 'foo/1.0' provide '['bar']'"




from unittest import mock
import pytest

from conan.internal.api.detect.detect_api import detect_gcc_compiler

@pytest.mark.parametrize("version", ["10", "4.2", '7', "8.12"])
def test_detect_gcc_10(version):
    with mock.patch("platform.system", return_value="Linux"):
        with mock.patch("conan.internal.api.detect.detect_api.detect_runner", return_value=(0, version)):
            compiler, installed_version, compiler_exe = detect_gcc_compiler()
    assert compiler == 'gcc'
    assert installed_version == version
    assert compiler_exe == 'gcc'



import os
from typing import List, Tuple

from conan import ConanFile
from conan.api.output import ConanOutput
from conan.cli.printers.graph import print_graph_basic
from conan.internal.cache.cache import PkgCache
from conan.internal.conan_app import ConanApp
from conan.internal.api.export import cmd_export
from conan.internal.methods import run_package_method
from conan.internal.graph.graph import BINARY_BUILD, RECIPE_INCACHE
from conan.api.model import PkgReference, Remote, RecipeReference
from conan.internal.util.files import mkdir


class ExportAPI:
    """ This API provides methods to export artifacts, both recipes and pre-compiled package
    binaries from user folders to the Conan cache, as Conan recipes and Conan package binaries
    """

    def __init__(self, conan_api, helpers):
        self._conan_api = conan_api
        self._helpers = helpers

    def export(self, path, name: str = None, version: str = None, user: str = None,
               channel: str = None, lockfile=None,
               remotes: List[Remote] = None) -> Tuple[RecipeReference, ConanFile]:
        """ Exports a ``conanfile.py`` recipe, together with its associated files to the Conan cache.
        A "recipe-revision" will be computed and assigned.

        :param path: Path to the conanfile to be exported
        :param name: Optional package name. Typically not necessary as it is defined by the recipe
            attribute or dynamically with the ``set_name()`` method.
            If it is defined in recipe and as an argument, but they don't match, an error will be raised.
        :param version: Optional version. It can be defined in the recipe with the version
            attribute or dynamically with the 'set_version()' method.
            If it is defined in recipe and as an argument, but they don't match, an error will be raised.
        :param user: Optional user. Can be defined by recipe attribute.
            If it is defined in recipe and as an argument, but they don't match, an error will be raised.
        :param channel: Optional channel. Can be defined by recipe attribute.
            If it is defined in recipe and as an argument, but they don't match, an error will be raised.
        :param lockfile: Optional, only relevant if the recipe has 'python-requires' to be locked
        :param remotes: Optional, only relevant to resolve 'python-requires' in remotes
        :return: A tuple of the exported RecipeReference and a ConanFile object
        """
        ConanOutput().title("Exporting recipe to the cache")
        app = ConanApp(self._conan_api)
        hook_manager = self._helpers.hook_manager
        return cmd_export(app.loader, app.cache, hook_manager, self._helpers.global_conf, path,
                          name, version, user, channel, graph_lock=lockfile, remotes=remotes)

    def export_pkg_graph(self, path, ref: RecipeReference, profile_host, profile_build,
                         remotes: List[Remote], lockfile=None, is_build_require=False,
                         skip_binaries=False, output_folder=None):
        """Computes a dependency graph for a given configuration, for an already existing (previously
        exported) recipe in the Conan cache. This method computes the full dependency graph, using
        the profiles, lockfile and remotes information as any other install/graph/create command.
        This is necessary in order to compute the "package_id" of the binary being exported
        into the Conan cache.
        The resulting dependency graph can be passed to ``export_pkg()`` method

        :param path: Path to the conanfile.py in the user folder
        :param ref: full RecipeReference, including recipe-revision
        :param profile_host: Profile for the host context
        :param profile_build: Profile for the build context
        :param lockfile: Optional lockfile
        :param remotes: List of Remotes
        :param is_build_require: In case a package intended to be used as a tool-requires
        :param skip_binaries:
        :param output_folder: The folder containing output files, like potential environment scripts
        :return: A Graph object that can be passed to ``export_pkg()`` method
        """
        assert ref.revision, "ref argument must have recipe-revision defined"
        conan_api = self._conan_api
        deps_graph = conan_api.graph.load_graph_consumer(path,
                                                         ref.name, str(ref.version), ref.user,
                                                         ref.channel,
                                                         profile_host=profile_host,
                                                         profile_build=profile_build,
                                                         lockfile=lockfile, remotes=remotes,
                                                         update=None,
                                                         is_build_require=is_build_require)

        print_graph_basic(deps_graph)
        deps_graph.report_graph_error()
        conan_api.graph.analyze_binaries(deps_graph, build_mode=[ref.name], lockfile=lockfile,
                                         remotes=remotes)
        deps_graph.report_graph_error()

        root_node = deps_graph.root
        root_node.ref = ref  # Make sure the root node revision is well defined

        if not skip_binaries:
            # unless the user explicitly opts-out with --skip-binaries, it is necessary to install
            # binaries, in case there are build_requires necessary like tool-requires=cmake
            # and package() method doing ``cmake.install()``
            # for most cases, deps will be in cache already because of a previous "conan install"
            # but if it is not the case, the binaries from remotes will be downloaded
            conan_api.install.install_binaries(deps_graph=deps_graph, remotes=remotes)
        source_folder = os.path.dirname(path)
        conan_api.install.install_consumer(deps_graph=deps_graph, source_folder=source_folder,
                                           output_folder=output_folder)
        return deps_graph

    def export_pkg(self, graph, output_folder=None) -> None:
        """Executes the ``package()`` method of the exported recipe in order to copy the artifacts
        from user folder to the Conan cache package folder

        :param graph: A Graph object
        :param output_folder: Optional folder where generated files like environment scripts
            of dependencies have been installed
        """
        cache = PkgCache(self._conan_api.cache_folder, self._helpers.global_conf)
        hook_manager = self._helpers.hook_manager

        # The graph has to be loaded with build_mode=[ref.name], so that node is not tried
        # to be downloaded from remotes
        # passing here the create_reference=ref argument is useful so the recipe is in "develop",
        # because the "package()" method is in develop=True already
        pkg_node = graph.root
        ref = pkg_node.ref
        source_folder = os.path.dirname(pkg_node.path)
        out = ConanOutput(scope=pkg_node.conanfile.display_name)
        out.info("Exporting binary from user folder to Conan cache")
        conanfile = pkg_node.conanfile

        package_id = pkg_node.package_id
        assert package_id is not None
        out.info("Packaging to %s" % package_id)
        pref = PkgReference(ref, package_id)
        pkg_layout = cache.create_build_pkg_layout(pref)

        conanfile.folders.set_base_folders(source_folder, output_folder)
        dest_package_folder = pkg_layout.package()
        conanfile.folders.set_base_package(dest_package_folder)
        mkdir(pkg_layout.metadata())
        conanfile.folders.set_base_pkg_metadata(pkg_layout.metadata())

        with pkg_layout.set_dirty_context_manager():
            prev = run_package_method(conanfile, package_id, hook_manager, ref)

        pref = PkgReference(pref.ref, pref.package_id, prev)
        pkg_layout.reference = pref
        cache.assign_prev(pkg_layout)
        pkg_node.prev = prev
        pkg_node.pref_timestamp = pref.timestamp  # assigned by assign_prev
        pkg_node.recipe = RECIPE_INCACHE
        pkg_node.binary = BINARY_BUILD
        # Make sure folder is updated
        final_folder = pkg_layout.package()
        conanfile.folders.set_base_package(final_folder)
        out.info(f"Package folder {final_folder}")
        out.success("Exported package binary")



import platform
import textwrap

import pytest

from conan.test.utils.tools import TestClient


def test_cmake_find_none_transitive():
    c = TestClient()

    qt = textwrap.dedent("""
        from conan import ConanFile
        from conan.tools.cmake import CMake

        class pkgRecipe(ConanFile):
            name = "qt"
            version = "0.1"
            package_type = "static-library"

            # Binary configuration
            settings = "os", "compiler", "build_type", "arch"
            generators = "CMakeToolchain"

            # Sources are located in the same place as this recipe, copy them to the recipe
            exports_sources = "CMakeLists.txt", "src/*", "include/*"

            def build(self):
                cmake = CMake(self)
                cmake.configure()
                cmake.build()

            def package(self):
                cmake = CMake(self)
                cmake.install()

            def package_info(self):
                self.cpp_info.builddirs = ["qt/cmake"]
                self.cpp_info.set_property("cmake_find_mode", "none")
                self.cpp_info.set_property("cmake_file_name", "Qt5")
        """)

    cmake = textwrap.dedent("""
        cmake_minimum_required(VERSION 3.15)
        project(MyHello NONE)

        add_library(qt INTERFACE)
        install(TARGETS qt EXPORT Qt5Config)
        export(TARGETS qt
            NAMESPACE qt::
            FILE "${CMAKE_CURRENT_BINARY_DIR}/Qt5Config.cmake"
        )
        install(EXPORT Qt5Config
            DESTINATION "qt/cmake"
            NAMESPACE qt::
        )
        """)
    c.save({"conanfile.py": qt,
            "CMakeLists.txt": cmake})
    c.run("create .")

    karchive = textwrap.dedent("""
        from conan import ConanFile
        from conan.tools.cmake import CMake

        class pkgRecipe(ConanFile):
            name = "karchive"
            version = "0.1"
            package_type = "static-library"

            requires = "qt/0.1"
        """)

    c.save({"conanfile.py": karchive}, clean_first=True)
    c.run("create .")

    consumer_cmake = textwrap.dedent("""
        cmake_minimum_required(VERSION 3.15)
        project(MyHello NONE)

        find_package(karchive CONFIG REQUIRED)
        """)

    consumer = textwrap.dedent("""
        from conan import ConanFile
        from conan.tools.cmake import CMake

        class pkgRecipe(ConanFile):
            package_type = "static-library"
            settings = "os", "compiler", "build_type", "arch"
            generators = "CMakeToolchain", "CMakeDeps"

            def requirements(self):
                self.requires("karchive/0.1")
                self.requires("qt/0.1")

            def build(self):
                cmake = CMake(self)
                cmake.configure()
                cmake.build()
        """)

    c.save({"conanfile.py": consumer,
            "CMakeLists.txt": consumer_cmake}, clean_first=True)
    c.run("build .")
    assert "Conan: Target declared 'karchive::karchive'" in c.out
    # And it doesn't fail to find transitive qt


def test_cmake_find_none_relocation():
    c = TestClient(default_server_user=True)
    c.run("new cmake_lib -d name=pkg -d version=0.1")
    conanfile = c.load("conanfile.py")
    conanfile = conanfile + '        self.cpp_info.set_property("cmake_find_mode", "none")'
    conanfile = conanfile + '\n        self.cpp_info.builddirs = ["pkg/cmake"]'

    cmake_export = textwrap.dedent("""
        cmake_minimum_required(VERSION 3.15)
        project(MyHello CXX)

        add_library(pkg src/pkg.cpp)
        target_include_directories(pkg PUBLIC
          $<BUILD_INTERFACE:${CMAKE_CURRENT_SOURCE_DIR}/include>
          $<INSTALL_INTERFACE:include>
        )
        set_target_properties(pkg PROPERTIES PUBLIC_HEADER "include/pkg.h")

        install(TARGETS pkg EXPORT pkgConfig)
        export(TARGETS pkg
            NAMESPACE pkg::
            FILE "${CMAKE_CURRENT_BINARY_DIR}/pkgConfig.cmake"
        )
        install(EXPORT pkgConfig
            DESTINATION "pkg/cmake"
            NAMESPACE pkg::
        )
        """)

    c.save({"conanfile.py": conanfile,
            "CMakeLists.txt": cmake_export})

    c.run("create . ")
    c.run("upload * -r=default -c")
    c.run("remove * -c")

    c2 = TestClient(servers=c.servers)
    c2.run("new cmake_exe -d name=myapp -d version=0.1 -d requires=pkg/0.1")
    c2.run('build .')
    # Now it builds correctly without failing, because package is relocatable


@pytest.mark.tool("ninja")
# It is IMPORTANT to do cmake-3.27 AFTER ninja, otherwise CI injects another cmake
@pytest.mark.tool("cmake", "3.27")
def test_cmake_find_none_relocation_multi():
    # What happens for multi-config
    c = TestClient(default_server_user=True)
    c.run("new cmake_lib -d name=pkg -d version=0.1")
    conanfile = textwrap.dedent("""
        import os, textwrap
        from conan import ConanFile
        from conan.tools.files import save, load
        from conan.tools.cmake import CMakeToolchain, CMake, cmake_layout, CMakeDeps

        class pkgRecipe(ConanFile):
            name = "pkg"
            version = "0.1"
            package_type = "library"

            settings = "os", "compiler", "build_type", "arch"
            options = {"shared": [True, False], "fPIC": [True, False]}
            default_options = {"shared": False, "fPIC": True}

            # Sources are located in the same place as this recipe, copy them to the recipe
            exports_sources = "CMakeLists.txt", "src/*", "include/*"
            implements = ["auto_shared_fpic"]

            generators = "CMakeToolchain", "CMakeDeps"

            def layout(self):
                cmake_layout(self)

            def build(self):
                cmake = CMake(self)
                cmake.configure()
                cmake.build()

            def package(self):
                cmake = CMake(self)
                cmake.install()
                p = os.path.join(self.package_folder, "pkg", "cmake", "pkgConfig.cmake")
                new_config = textwrap.dedent('''
                    # Create imported target pkg::pkg
                    add_library(pkg::pkg STATIC IMPORTED)

                    if(NOT DEFINED CONAN_pkg_DIR_MULTI)
                        get_filename_component(CONAN_pkg_DIR_MULTI "${CMAKE_CURRENT_LIST_FILE}" PATH)
                    endif()
                    foreach(_IMPORT_PREFIX ${CONAN_pkg_DIR_MULTI})
                        get_filename_component(_IMPORT_PREFIX "${_IMPORT_PREFIX}" PATH)
                        get_filename_component(_IMPORT_PREFIX "${_IMPORT_PREFIX}" PATH)

                        set_target_properties(pkg::pkg PROPERTIES
                          INTERFACE_INCLUDE_DIRECTORIES "${_IMPORT_PREFIX}/include"
                        )

                        # Load information for each installed configuration.
                        file(GLOB CONFIG_FILES "${_IMPORT_PREFIX}/pkg/cmake/pkgConfig-*.cmake")
                        foreach(f ${CONFIG_FILES})
                          include(${f})
                        endforeach()
                    endforeach()
                    ''')
                save(self, p, new_config)

            def package_info(self):
                self.cpp_info.libs = ["pkg"]
                self.cpp_info.set_property("cmake_find_mode", "none")
                self.cpp_info.builddirs = ["pkg/cmake"]
        """)

    cmake_export = textwrap.dedent("""
        cmake_minimum_required(VERSION 3.15)
        project(MyHello CXX)

        add_library(pkg src/pkg.cpp)
        target_include_directories(pkg PUBLIC
          $<BUILD_INTERFACE:${CMAKE_CURRENT_SOURCE_DIR}/include>
          $<INSTALL_INTERFACE:include>
        )
        set_target_properties(pkg PROPERTIES PUBLIC_HEADER "include/pkg.h")

        install(TARGETS pkg EXPORT pkgConfig)
        export(TARGETS pkg
            NAMESPACE pkg::
            FILE "${CMAKE_CURRENT_BINARY_DIR}/pkgConfig.cmake"
        )
        install(EXPORT pkgConfig
            DESTINATION "pkg/cmake"
            NAMESPACE pkg::
        )
        """)

    c.save({"conanfile.py": conanfile,
            "CMakeLists.txt": cmake_export})

    c.run("create .")
    c.run("create . -s build_type=Debug")
    c.run("upload * -r=default -c")
    c.run("remove * -c")

    c2 = TestClient(servers=c.servers)
    c2.run("new cmake_exe -d name=myapp -d version=0.1 -d requires=pkg/0.1")
    env = r".\build\generators\conanbuild.bat &&" if platform.system() == "Windows" else ""
    conf = ('-c tools.cmake.cmaketoolchain:generator="Ninja Multi-Config" '
            '-c tools.cmake.cmakedeps:new=will_break_next')
    c2.run(f'install . {conf}')
    c2.run(f'install . {conf} -s build_type=Debug')
    c2.run_command(f"{env} cmake --preset conan-default")
    c2.run_command(f"{env} cmake --build --preset conan-release")
    c2.run_command(f"{env} cmake --build --preset conan-debug")



import json
import os
import platform
import shutil
import tarfile
import time

import pytest

from conan.api.model import PkgReference, RecipeReference
from conan.test.assets.genconanfile import GenConanfile
from conan.test.utils.test_files import temp_folder
from conan.test.utils.tools import TestClient, NO_SETTINGS_PACKAGE_ID
from conan.internal.util.files import save, load



def test_cache_save_restore():
    c = TestClient()
    c.save({"conanfile.py": GenConanfile().with_settings("os")})
    c.run("create . --name=pkg --version=1.0 -s os=Linux")
    c.run("create . --name=pkg --version=1.1 -s os=Linux")
    c.run("create . --name=other --version=2.0 -s os=Linux")
    # Force the compress level just to make sure it doesn't crash
    c.run("cache save pkg/*:* -cc core.gzip:compresslevel=9")
    cache_path = os.path.join(c.current_folder, "conan_cache_save.tgz")
    assert os.path.exists(cache_path)
    _validate_restore(cache_path)

    # Lets test that the pkglist does not contain windows backslash paths to make it portable
    with open(cache_path, mode='rb') as file_handler:
        the_tar = tarfile.open(fileobj=file_handler)
        fileobj = the_tar.extractfile("pkglist.json")
        pkglist = fileobj.read()
        the_tar.close()

    package_list = json.loads(pkglist)
    assert "\\" not in package_list


def test_cache_save_restore_with_package_file():
    """If we have some sources in the root (like the CMakeLists.txt)
    we don't declare folders.source"""
    conan_file = GenConanfile() \
        .with_settings("os") \
        .with_package_file("bin/file.txt", "content!!")

    client = TestClient()
    client.save({"conanfile.py": conan_file})
    client.run("create . --name=pkg --version=1.0 -s os=Linux")
    client.run("cache save pkg/*:* ")
    cache_path = os.path.join(client.current_folder, "conan_cache_save.tgz")
    assert os.path.exists(cache_path)

    c2 = TestClient()
    shutil.copy2(cache_path, c2.current_folder)
    c2.run("cache restore conan_cache_save.tgz")
    c2.run("list *:*#*")
    assert "pkg/1.0" in c2.out
    tree = _get_directory_tree(c2.base_folder)

    # Restore again, expect the tree to be unchanged
    c2.run("cache restore conan_cache_save.tgz")
    c2.run("list *:*#*")
    assert "pkg/1.0" in c2.out
    tree2 = _get_directory_tree(c2.base_folder)

    assert tree2 == tree


def test_cache_save_downloaded_restore():
    """ what happens if we save packages downloaded from server, not
    created
    """
    c = TestClient(default_server_user=True)
    c.save({"conanfile.py": GenConanfile().with_settings("os")})
    c.run("create . --name=pkg --version=1.0 -s os=Linux")
    c.run("create . --name=pkg --version=1.1 -s os=Linux")
    c.run("create . --name=other --version=2.0 -s os=Linux")
    c.run("upload * -r=default -c")
    c.run("remove * -c")
    c.run("download *:* -r=default --metadata=*")
    c.run("cache save pkg/*:* ")
    cache_path = os.path.join(c.current_folder, "conan_cache_save.tgz")
    assert os.path.exists(cache_path)

    _validate_restore(cache_path)


def _get_directory_tree(base_folder):
    tree = []
    for d, _, fs in os.walk(base_folder):
        rel_d = os.path.relpath(d, base_folder) if d != base_folder else ""
        if rel_d:
            tree.append(rel_d)
        for f in fs:
            tree.append(os.path.join(rel_d, f))
    tree.sort()
    return tree


def _validate_restore(cache_path):
    c2 = TestClient()
    # Create a package in the cache to check put doesn't interact badly
    c2.save({"conanfile.py": GenConanfile().with_settings("os")})
    c2.run("create . --name=pkg2 --version=3.0 -s os=Windows")
    shutil.copy2(cache_path, c2.current_folder)
    c2.run("cache restore conan_cache_save.tgz")
    c2.run("list *:*#*")
    assert "pkg2/3.0" in c2.out
    assert "pkg/1.0" in c2.out
    assert "pkg/1.1" in c2.out
    assert "other/2.0" not in c2.out
    tree = _get_directory_tree(c2.base_folder)

    # Restore again, just in case
    c2.run("cache restore conan_cache_save.tgz")
    c2.run("list *:*#*")
    assert "pkg2/3.0" in c2.out
    assert "pkg/1.0" in c2.out
    assert "pkg/1.1" in c2.out
    assert "other/2.0" not in c2.out
    tree2 = _get_directory_tree(c2.base_folder)
    assert tree2 == tree


def test_cache_save_excluded_folders():
    # https://github.com/conan-io/conan/issues/18234
    c = TestClient(default_server_user=True)
    c.save({"conanfile.py": GenConanfile().with_exports("*.py").with_exports_sources("*.c"),
            "somefile.py": "",
            "mysrc.c": ""})
    c.run("create . --name=pkg --version=1.0")
    ref_layout = c.exported_layout()
    pkg_layout = c.created_layout()
    c.run("upload * --dry-run -r=default -c")
    assert os.path.exists(os.path.join(ref_layout.download_export(), "conan_export.tgz"))
    assert os.path.exists(os.path.join(ref_layout.download_export(), "conan_sources.tgz"))
    assert os.path.exists(os.path.join(ref_layout.source(), "mysrc.c"))
    assert os.path.exists(os.path.join(pkg_layout.download_package(), "conan_package.tgz"))
    assert os.path.exists(pkg_layout.build())

    c.run("cache save *:*")
    cache_path = os.path.join(c.current_folder, "conan_cache_save.tgz")

    c2 = TestClient()
    shutil.copy2(cache_path, c2.current_folder)
    c2.run("cache restore conan_cache_save.tgz")

    ref = RecipeReference.loads("pkg/1.0")
    ref_layout = c2.get_latest_ref_layout(ref)
    pkg_layout = c2.get_latest_pkg_layout(PkgReference(ref_layout.reference, NO_SETTINGS_PACKAGE_ID))
    assert os.path.exists(os.path.join(ref_layout.source(), "mysrc.c"))
    assert not os.path.exists(os.path.join(ref_layout.download_export(), "conan_export.tgz"))
    assert not os.path.exists(os.path.join(ref_layout.download_export(), "conan_sources.tgz"))
    assert not os.path.exists(os.path.join(pkg_layout.download_package(), "conan_package.tgz"))
    assert not os.path.exists(pkg_layout.build())

    # exclude source
    c.run("cache save * --no-source")
    c3 = TestClient()
    shutil.copy2(cache_path, c3.current_folder)
    c3.run("cache restore conan_cache_save.tgz")
    ref_layout = c3.get_latest_ref_layout(ref)
    assert not os.path.exists(os.path.join(ref_layout.source(), "mysrc.c"))


def test_cache_save_restore_metadata():
    c = TestClient()
    c.save({"conanfile.py": GenConanfile().with_settings("os")})
    c.run("create . --name=pkg --version=1.0 -s os=Linux")
    pid = c.created_package_id("pkg/1.0")
    # Add some metadata
    c.run("cache path pkg/1.0 --folder=metadata")
    metadata_path = str(c.stdout).strip()
    myfile = os.path.join(metadata_path, "logs", "mylogs.txt")
    save(myfile, "mylogs!!!!")
    c.run(f"cache path pkg/1.0:{pid} --folder=metadata")
    pkg_metadata_path = str(c.stdout).strip()
    myfile = os.path.join(pkg_metadata_path, "logs", "mybuildlogs.txt")
    save(myfile, "mybuildlogs!!!!")

    c.run("cache save  pkg/*:* ")
    cache_path = os.path.join(c.current_folder, "conan_cache_save.tgz")
    assert os.path.exists(cache_path)

    # restore and check
    c2 = TestClient()
    shutil.copy2(cache_path, c2.current_folder)
    c2.run("cache restore conan_cache_save.tgz")
    c2.run("cache path pkg/1.0 --folder=metadata")
    metadata_path = str(c2.stdout).strip()
    myfile = os.path.join(metadata_path, "logs", "mylogs.txt")
    assert load(myfile) == "mylogs!!!!"
    c2.run(f"cache path pkg/1.0:{pid} --folder=metadata")
    pkg_metadata_path = str(c2.stdout).strip()
    myfile = os.path.join(pkg_metadata_path, "logs", "mybuildlogs.txt")
    assert load(myfile) == "mybuildlogs!!!!"


# FIXME: check the timestamps of the conan cache restore
@pytest.mark.skipif(platform.system() == "Windows", reason="Fails in windows in ci because of the low precission of the clock")
def test_cache_save_restore_multiple_revisions():
    c = TestClient()
    c.save({"conanfile.py": GenConanfile("pkg", "0.1")})
    c.run("create .")
    rrev1 = c.exported_recipe_revision()
    time.sleep(0.2)
    c.save({"conanfile.py": GenConanfile("pkg", "0.1").with_class_attribute("var=42")})
    c.run("create .")
    rrev2 = c.exported_recipe_revision()
    time.sleep(0.2)
    c.save({"conanfile.py": GenConanfile("pkg", "0.1").with_class_attribute("var=123")})
    c.run("create .")
    rrev3 = c.exported_recipe_revision()


    def check_ordered_revisions(client):
        client.run("list *#* --format=json")
        revisions = json.loads(client.stdout)["Local Cache"]["pkg/0.1"]["revisions"]
        assert revisions[rrev1]["timestamp"] < revisions[rrev2]["timestamp"]
        assert revisions[rrev2]["timestamp"] < revisions[rrev3]["timestamp"]

    check_ordered_revisions(c)

    c.run("cache save pkg/*#*:* ")
    cache_path = os.path.join(c.current_folder, "conan_cache_save.tgz")

    # restore and check
    c2 = TestClient()
    shutil.copy2(cache_path, c2.current_folder)
    c2.run("cache restore conan_cache_save.tgz")
    check_ordered_revisions(c2)


def test_cache_save_restore_graph():
    """ It is possible to save package list
    """
    c = TestClient()
    c.save({"dep/conanfile.py": GenConanfile("dep", "0.1"),
            "pkg/conanfile.py": GenConanfile("pkg", "0.1").with_requires("dep/0.1")})
    c.run("create dep")
    c.run("create pkg --format=json", redirect_stdout="graph.json")
    c.run("list --graph=graph.json --format=json", redirect_stdout="list.json")
    c.run("cache save --file=cache.tgz --list=list.json")
    cache_path = os.path.join(c.current_folder, "cache.tgz")
    assert os.path.exists(cache_path)
    c2 = TestClient()
    # Create a package in the cache to check put doesn't interact badly
    c2.save({"conanfile.py": GenConanfile().with_settings("os")})
    c2.run("create . --name=pkg2 --version=3.0 -s os=Windows")
    shutil.copy2(cache_path, c2.current_folder)
    c2.run("cache restore cache.tgz")
    c2.run("list *:*#*")
    assert "pkg/0.1" in c2.out
    assert "dep/0.1" in c2.out


def test_cache_save_subfolder():
    """ It is possible to save package list in subfolder that doesn't exist
    https://github.com/conan-io/conan/issues/15362
    """
    c = TestClient()
    c.save({"conanfile.py": GenConanfile("dep", "0.1")})
    c.run("export .")
    c.run("cache save * --file=subfolder/cache.tgz")
    assert os.path.exists(os.path.join(c.current_folder, "subfolder", "cache.tgz"))


def test_error_restore_not_existing():
    c = TestClient()
    c.run("cache restore potato.tgz", assert_error=True)
    assert "ERROR: Restore archive doesn't exist in " in c.out


@pytest.mark.parametrize("src_store", (False, True))
@pytest.mark.parametrize("dst_store", (False, True))
def test_cache_save_restore_custom_storage_path(src_store, dst_store):
    c = TestClient()
    if src_store:
        tmp_folder = temp_folder()
        c.save_home({"global.conf": f"core.cache:storage_path={tmp_folder}"})
    c.save({"conanfile.py": GenConanfile()})
    c.run("create . --name=pkg --version=1.0")
    c.run("cache save *:*")
    cache_path = os.path.join(c.current_folder, "conan_cache_save.tgz")

    c2 = TestClient()
    if dst_store:
        tmp_folder = temp_folder()
        c2.save_home({"global.conf": f"core.cache:storage_path={tmp_folder}"})
    shutil.copy2(cache_path, c2.current_folder)
    c2.run("cache restore conan_cache_save.tgz")
    c2.run("list *:*")
    assert "pkg/1.0" in c2.out



import os
import platform
import shutil
import textwrap

import pytest

from conan.test.assets.cmake import gen_cmakelists
from conan.test.assets.genconanfile import GenConanfile
from conan.test.assets.sources import gen_function_cpp
from conan.test.utils.test_files import temp_folder
from conan.test.utils.tools import TestClient
from conan.internal.util.files import save


@pytest.fixture()
def client(matrix_client_shared):
    c = matrix_client_shared
    conanfile = textwrap.dedent("""
       import os
       from conan import ConanFile
       from conan.tools.files import save
       class Tool(ConanFile):
           name = "tool"
           version = "1.0"
           def package(self):
               save(self, os.path.join(self.package_folder, "build", "my_tools.cmake"),
                    'set(MY_TOOL_VARIABLE "Hello world!")')

           def package_info(self):
               self.cpp_info.includedirs = []
               self.cpp_info.libdirs = []
               self.cpp_info.bindirs = []
               path_build_modules = os.path.join("build", "my_tools.cmake")
               self.cpp_info.set_property("cmake_build_modules", [path_build_modules])
           """)
    c.save({"conanfile.py": conanfile}, clean_first=True)
    c.run("create .")
    return c


@pytest.mark.tool("cmake")
@pytest.mark.parametrize("powershell", [False, True])
def test_install_deploy(client, powershell):
    c = client
    custom_content = 'message(STATUS "MY_TOOL_VARIABLE=${MY_TOOL_VARIABLE}!")'
    cmake = gen_cmakelists(appname="my_app", appsources=["main.cpp"], find_package=["matrix", "tool"],
                           custom_content=custom_content)
    deploy = textwrap.dedent("""
        import os, shutil

        # USE **KWARGS to be robust against changes
        def deploy(graph, output_folder, **kwargs):
            conanfile = graph.root.conanfile
            for r, d in conanfile.dependencies.items():
                new_folder = os.path.join(output_folder, d.ref.name)
                shutil.copytree(d.package_folder, new_folder)
                d.set_deploy_folder(new_folder)
        """)
    c.save({"conanfile.txt": "[requires]\nmatrix/1.0\ntool/1.0",
            "deploy.py": deploy,
            "CMakeLists.txt": cmake,
            "main.cpp": gen_function_cpp(name="main", includes=["matrix"], calls=["matrix"])},
           clean_first=True)
    pwsh = "-c tools.env.virtualenv:powershell=True" if powershell else ""
    c.run("install . -o *:shared=True "
          f"--deployer=deploy.py -of=mydeploy -g CMakeToolchain -g CMakeDeps {pwsh}")
    c.run("remove * -c")  # Make sure the cache is clean, no deps there
    arch = c.get_default_host_profile().settings['arch']
    deps = c.load(f"mydeploy/matrix-release-{arch}-data.cmake")
    assert 'set(matrix_PACKAGE_FOLDER_RELEASE "${CMAKE_CURRENT_LIST_DIR}/matrix")' in deps
    assert 'set(matrix_INCLUDE_DIRS_RELEASE "${matrix_PACKAGE_FOLDER_RELEASE}/include")' in deps
    assert 'set(matrix_LIB_DIRS_RELEASE "${matrix_PACKAGE_FOLDER_RELEASE}/lib")' in deps

    # We can fully move it to another folder, and still works
    tmp = os.path.join(temp_folder(), "relocated")
    shutil.copytree(c.current_folder, tmp)
    shutil.rmtree(c.current_folder)
    c2 = TestClient(current_folder=tmp)
    # I can totally build without errors with deployed
    c2.run_command("cmake . -DCMAKE_TOOLCHAIN_FILE=mydeploy/conan_toolchain.cmake "
                   "-DCMAKE_BUILD_TYPE=Release")
    assert "MY_TOOL_VARIABLE=Hello world!!" in c2.out
    c2.run_command("cmake --build . --config Release")
    if platform.system() == "Windows":  # Only the .bat env-generators are relocatable
        if powershell:
            cmd = r"powershell.exe mydeploy\conanrun.ps1 ; Release\my_app.exe"
        else:
            cmd = r"mydeploy\conanrun.bat && Release\my_app.exe"
        # For Lunux: cmd = ". mydeploy/conanrun.sh && ./my_app"
        c2.run_command(cmd)
        assert "matrix/1.0: Hello World Release!" in c2.out


@pytest.mark.tool("cmake")
def test_install_full_deploy_layout(client):
    c = client
    custom_content = 'message(STATUS "MY_TOOL_VARIABLE=${MY_TOOL_VARIABLE}!")'
    cmake = gen_cmakelists(appname="my_app", appsources=["main.cpp"], find_package=["matrix", "tool"],
                           custom_content=custom_content)
    conanfile = textwrap.dedent("""
        [requires]
        matrix/1.0
        tool/1.0
        [generators]
        CMakeDeps
        CMakeToolchain
        [layout]
        cmake_layout
        """)
    c.save({"conanfile.txt": conanfile,
            "CMakeLists.txt": cmake,
            "main.cpp": gen_function_cpp(name="main", includes=["matrix"], calls=["matrix"])},
           clean_first=True)
    c.run("install . -o *:shared=True --deployer=full_deploy.py")
    c.run("remove * -c")  # Make sure the cache is clean, no deps there
    arch = c.get_default_host_profile().settings['arch']
    folder = "/Release" if platform.system() != "Windows" else ""
    rel_path = "../../" if platform.system() == "Windows" else "../../../"
    deps = c.load(f"build{folder}/generators/matrix-release-{arch}-data.cmake")
    assert 'set(matrix_PACKAGE_FOLDER_RELEASE "${CMAKE_CURRENT_LIST_DIR}/' \
           f'{rel_path}full_deploy/host/matrix/1.0/Release/{arch}")' in deps
    assert 'set(matrix_INCLUDE_DIRS_RELEASE "${matrix_PACKAGE_FOLDER_RELEASE}/include")' in deps
    assert 'set(matrix_LIB_DIRS_RELEASE "${matrix_PACKAGE_FOLDER_RELEASE}/lib")' in deps

    # We can fully move it to another folder, and still works
    tmp = os.path.join(temp_folder(), "relocated")
    shutil.copytree(c.current_folder, tmp)
    shutil.rmtree(c.current_folder)
    c2 = TestClient(current_folder=tmp)
    with c2.chdir(f"build{folder}"):
        # I can totally build without errors with deployed
        cmakelist = "../.." if platform.system() != "Windows" else ".."
        c2.run_command(f"cmake {cmakelist} -DCMAKE_TOOLCHAIN_FILE=generators/conan_toolchain.cmake "
                       "-DCMAKE_BUILD_TYPE=Release")
        assert "MY_TOOL_VARIABLE=Hello world!!" in c2.out
        c2.run_command("cmake --build . --config Release")
        if platform.system() == "Windows":  # Only the .bat env-generators are relocatable atm
            cmd = r"generators\conanrun.bat && Release\my_app.exe"
            # For Lunux: cmd = ". mydeploy/conanrun.sh && ./my_app"
            c2.run_command(cmd)
            assert "matrix/1.0: Hello World Release!" in c2.out


def test_copy_files_deploy():
    c = TestClient()
    deploy = textwrap.dedent("""
        import os, shutil

        def deploy(graph, output_folder, **kwargs):
            conanfile = graph.root.conanfile
            for r, d in conanfile.dependencies.items():
                bindir = os.path.join(d.package_folder, "bin")
                for f in os.listdir(bindir):
                    shutil.copy2(os.path.join(bindir, f), os.path.join(output_folder, f))
        """)
    c.save({"conanfile.txt": "[requires]\nhello/0.1",
            "deploy.py": deploy,
            "hello/conanfile.py": GenConanfile("hello", "0.1").with_package_file("bin/file.txt",
                                                                                 "content!!")})
    c.run("create hello")
    c.run("install . --deployer=deploy.py -of=mydeploy")


def test_multi_deploy():
    """ check that we can add more than 1 deployer in the command line, both in local folders
    and in cache.
    Also testing that using .py extension or not, is the same
    Also, the local folder have precedence over the cache extensions
    """
    c = TestClient()
    deploy1 = textwrap.dedent("""
        def deploy(graph, output_folder, **kwargs):
            conanfile = graph.root.conanfile
            conanfile.output.info("deploy1!!")
        """)
    deploy2 = textwrap.dedent("""
        def deploy(graph, output_folder, **kwargs):
            conanfile = graph.root.conanfile
            conanfile.output.info("sub/deploy2!!")
        """)
    deploy_cache = textwrap.dedent("""
        def deploy(graph, output_folder, **kwargs):
            conanfile = graph.root.conanfile
            conanfile.output.info("deploy cache!!")
        """)
    save(os.path.join(c.cache_folder, "extensions", "deploy", "deploy_cache.py"), deploy_cache)
    # This should never be called in this test, always the local is found first
    save(os.path.join(c.cache_folder, "extensions", "deploy", "mydeploy.py"), "CRASH!!!!")
    c.save({"conanfile.txt": "",
            "mydeploy.py": deploy1,
            "sub/mydeploy2.py": deploy2})

    c.run("install . --deployer=mydeploy --deployer=sub/mydeploy2 --deployer=deploy_cache")
    assert "conanfile.txt: deploy1!!" in c.out
    assert "conanfile.txt: sub/deploy2!!" in c.out
    assert "conanfile.txt: deploy cache!!" in c.out

    # Now with .py extension
    c.run("install . --deployer=mydeploy.py --deployer=sub/mydeploy2.py --deployer=deploy_cache.py")
    assert "conanfile.txt: deploy1!!" in c.out
    assert "conanfile.txt: sub/deploy2!!" in c.out
    assert "conanfile.txt: deploy cache!!" in c.out


def test_deploy_local_import():
    """ test that deployers can share some Python code with local imports
    """
    c = TestClient()
    helper = textwrap.dedent("""
        def myhelper(conanfile):
            conanfile.output.info("My HELPER!!")
        """)
    deploy_cache = textwrap.dedent("""
        from helper import myhelper
        def deploy(graph, output_folder, **kwargs):
            myhelper(graph.root.conanfile)
        """)
    save(os.path.join(c.cache_folder, "extensions", "deployers", "deploy_cache.py"), deploy_cache)
    save(os.path.join(c.cache_folder, "extensions", "deployers", "helper.py"), helper)
    c.save({"conanfile.txt": ""})
    c.run("install . --deployer=deploy_cache")
    assert "conanfile.txt: My HELPER!!" in c.out


def test_builtin_full_deploy():
    """ check the built-in full_deploy
    """
    c = TestClient()
    conanfile = textwrap.dedent("""
        import os
        from conan import ConanFile
        from conan.tools.files import save
        class Pkg(ConanFile):
            settings = "arch", "build_type"
            def package(self):
                content = f"{self.settings.build_type}-{self.settings.arch}"
                save(self, os.path.join(self.package_folder, "include/hello.h"), content)

            def package_info(self):
                path_build_modules = os.path.join("build", "my_tools_{}.cmake".format(self.context))
                self.cpp_info.set_property("cmake_build_modules", [path_build_modules])
            """)
    c.save({"conanfile.py": conanfile})
    c.run("create . --name=dep --version=0.1")
    c.run("create . --name=dep --version=0.1 -s build_type=Debug -s arch=x86")
    c.save({"conanfile.txt": "[requires]\ndep/0.1"}, clean_first=True)
    c.run("install . --deployer=full_deploy -of=output -g CMakeDeps")
    assert "Conan built-in full deployer" in c.out
    c.run("install . --deployer=full_deploy -of=output -g CMakeDeps "
          "-s build_type=Debug -s arch=x86")

    host_arch = c.get_default_host_profile().settings['arch']
    release = c.load(f"output/full_deploy/host/dep/0.1/Release/{host_arch}/include/hello.h")
    assert f"Release-{host_arch}" in release
    debug = c.load("output/full_deploy/host/dep/0.1/Debug/x86/include/hello.h")
    assert "Debug-x86" in debug
    cmake_release = c.load(f"output/dep-release-{host_arch}-data.cmake")
    assert 'set(dep_INCLUDE_DIRS_RELEASE "${dep_PACKAGE_FOLDER_RELEASE}/include")' in cmake_release
    assert f"${{CMAKE_CURRENT_LIST_DIR}}/full_deploy/host/dep/0.1/Release/{host_arch}" in cmake_release
    assert 'set(dep_BUILD_MODULES_PATHS_RELEASE ' \
           '"${dep_PACKAGE_FOLDER_RELEASE}/build/my_tools_host.cmake")' in cmake_release
    cmake_debug = c.load("output/dep-debug-x86-data.cmake")
    assert 'set(dep_INCLUDE_DIRS_DEBUG "${dep_PACKAGE_FOLDER_DEBUG}/include")' in cmake_debug
    assert "${CMAKE_CURRENT_LIST_DIR}/full_deploy/host/dep/0.1/Debug/x86" in cmake_debug
    assert 'set(dep_BUILD_MODULES_PATHS_DEBUG ' \
           '"${dep_PACKAGE_FOLDER_DEBUG}/build/my_tools_host.cmake")' in cmake_debug


def test_deploy_reference():
    """ check that we can also deploy a reference
    """
    c = TestClient()
    c.save({"conanfile.py": GenConanfile("pkg", "1.0").with_package_file("include/hi.h", "hi")})
    c.run("create .")

    c.run("install  --requires=pkg/1.0 --deployer=full_deploy --output-folder=output")
    # NOTE: Full deployer always use build_type/arch, even if None/None in the path, same structure
    header = c.load("output/full_deploy/host/pkg/1.0/include/hi.h")
    assert "hi" in header

    # Testing that we can deploy to the current folder too
    c.save({}, clean_first=True)
    c.run("install  --requires=pkg/1.0 --deployer=full_deploy")
    # NOTE: Full deployer always use build_type/arch, even if None/None in the path, same structure
    header = c.load("full_deploy/host/pkg/1.0/include/hi.h")
    assert "hi" in header


def test_deploy_overwrite():
    """ calling several times the install --deploy doesn't crash if files already exist
    """
    c = TestClient()
    c.save({"conanfile.py": GenConanfile("pkg", "1.0").with_package_file("include/hi.h", "hi")})
    c.run("create .")

    c.run("install  --requires=pkg/1.0 --deployer=full_deploy --output-folder=output")
    header = c.load("output/full_deploy/host/pkg/1.0/include/hi.h")
    assert "hi" in header

    # modify the package
    c.save({"conanfile.py": GenConanfile("pkg", "1.0").with_package_file("include/hi.h", "bye")})
    c.run("create .")
    c.run("install  --requires=pkg/1.0 --deployer=full_deploy --output-folder=output")
    header = c.load("output/full_deploy/host/pkg/1.0/include/hi.h")
    assert "bye" in header


def test_deploy_editable():
    """ when deploying something that is editable, with the full_deploy built-in, it will copy the
    editable files as-is, but it doesn't fail at this moment
    """

    c = TestClient()
    c.save({"conanfile.py": GenConanfile("pkg", "1.0"),
            "src/include/hi.h": "hi"})
    c.run("editable add .")

    # If we don't change to another folder, the full_deploy will be recursive and fail
    with c.chdir(temp_folder()):
        c.run("install  --requires=pkg/1.0 --deployer=full_deploy --output-folder=output")
        header = c.load("output/full_deploy/host/pkg/1.0/src/include/hi.h")
        assert "hi" in header


def test_deploy_aggregate_components():
    """
    The caching of aggregated components can be causing issues when deploying and using
    generators that would still point to the packages with components in the cache
    https://github.com/conan-io/conan/issues/14022
    """
    c = TestClient()
    dep = textwrap.dedent("""
        from conan import ConanFile
        class Pkg(ConanFile):
            name = "dep"
            version = "0.1"

            def package_info(self):
                self.cpp_info.components["mycomp"].libs = ["mycomp"]
            """)
    c.save({"dep/conanfile.py": dep,
            "pkg/conanfile.py": GenConanfile("pkg", "0.1").with_settings("build_type")
                                                          .with_requires("dep/0.1")
                                                          .with_generator("CMakeDeps"),
            "consumer/conanfile.py": GenConanfile().with_settings("build_type")
                                                   .with_requires("pkg/0.1")
                                                   .with_generator("CMakeDeps")})
    c.run("export dep")
    c.run("export pkg")

    # If we don't change to another folder, the full_deploy will be recursive and fail
    c.run("install consumer --build=missing --deployer=full_deploy --output-folder=output")
    data = c.load("output/dep-release-data.cmake")
    assert 'set(dep_PACKAGE_FOLDER_RELEASE ' \
           '"${CMAKE_CURRENT_LIST_DIR}/full_deploy/host/dep/0.1")' in data
    assert 'set(dep_INCLUDE_DIRS_RELEASE "${dep_PACKAGE_FOLDER_RELEASE}/include")' in data


def test_deploy_single_package():
    """ Let's try a deploy that executes on a single package reference
    """
    c = TestClient()
    c.save({"conanfile.py": GenConanfile("pkg", "1.0").with_package_file("include/hi.h", "hi"),
            "consumer/conanfile.txt": "[requires]\npkg/1.0"})
    c.run("create .")

    # if we deploy one --requires, we get that package
    c.run("install  --requires=pkg/1.0 --deployer=direct_deploy --output-folder=output")
    header = c.load("output/direct_deploy/pkg/include/hi.h")
    assert "hi" in header

    # If we deploy a local conanfile.txt, we get deployed its direct dependencies
    c.run("install consumer/conanfile.txt --deployer=direct_deploy --output-folder=output2")
    header = c.load("output2/direct_deploy/pkg/include/hi.h")
    assert "hi" in header


def test_deploy_output_locations():
    tc = TestClient()
    deployer = textwrap.dedent("""
    def deploy(graph, output_folder, **kwargs):
        graph.root.conanfile.output.info(f"Deployer output: {output_folder}")
    """)
    tc.save({"conanfile.txt": "",
             "my_deploy.py": deployer})

    tmp_folder = temp_folder()
    tc.run(f"install . --deployer=my_deploy -of='{tmp_folder}'")
    assert f"Deployer output: {tmp_folder}" in tc.out

    deployer_output = temp_folder()
    tc.run(f"install . --deployer=my_deploy -of='{tmp_folder}' --deployer-folder='{deployer_output}'")
    assert f"Deployer output: {deployer_output}" in tc.out
    assert f"Deployer output: {tmp_folder}" not in tc.out


def test_not_deploy_absolute_paths():
    """ Absolute paths, for system packages, don't need to be relativized
    https://github.com/conan-io/conan/issues/15242
    """
    c = TestClient()
    some_abs_path = temp_folder().replace("\\", "/")
    conanfile = textwrap.dedent(f"""
        from conan import ConanFile
        class Pkg(ConanFile):
            name = "pkg"
            version = "1.0"
            def package_info(self):
                self.cpp_info.includedirs = ["{some_abs_path}/myusr/include"]
                self.cpp_info.libdirs = ["{some_abs_path}/myusr/lib"]
                self.buildenv_info.define_path("MYPATH", "{some_abs_path}/mypath")
        """)
    c.save({"conanfile.py": conanfile})
    c.run("create .")

    # if we deploy one --requires, we get that package
    c.run("install  --requires=pkg/1.0 --deployer=full_deploy -g CMakeDeps -g CMakeToolchain "
          "-s os=Linux -s:b os=Linux -s arch=x86_64 -s:b arch=x86_64")
    data = c.load("pkg-release-x86_64-data.cmake")
    assert f'set(pkg_INCLUDE_DIRS_RELEASE "{some_abs_path}/myusr/include")' in data
    assert f'set(pkg_LIB_DIRS_RELEASE "{some_abs_path}/myusr/lib")' in data

    env = c.load("conanbuildenv-release-x86_64.sh")
    assert f'export MYPATH="{some_abs_path}/mypath"' in env


def test_deploy_incorrect_folder():
    # https://github.com/conan-io/cmake-conan/issues/658
    c = TestClient()
    c.save({"conanfile.txt": ""})
    c.run('install . --deployer=full_deploy --deployer-folder="mydep fold"')
    assert os.path.exists(os.path.join(c.current_folder, "mydep fold"))
    if platform.system() == "Windows":  # This only fails in Windows
        c.run(r'install . --deployer=full_deploy --deployer-folder="\"mydep fold\""',
              assert_error=True)
        assert "ERROR: Deployer folder cannot be created" in c.out


class TestRuntimeDeployer:
    def test_runtime_deploy(self):
        c = TestClient()
        conanfile = textwrap.dedent("""
           from conan import ConanFile
           from conan.tools.files import copy
           class Pkg(ConanFile):
               package_type = "shared-library"
               def package(self):
                   copy(self, "*.so", src=self.build_folder, dst=self.package_folder)
                   copy(self, "*.dll", src=self.build_folder, dst=self.package_folder)
           """)
        c.save({"pkga/conanfile.py": conanfile,
                "pkga/lib/pkga.so": "",
                "pkga/bin/pkga.dll": "",
                "pkgb/conanfile.py": conanfile,
                "pkgb/lib/pkgb.so": ""})
        c.run("export-pkg pkga --name=pkga --version=1.0")
        c.run("export-pkg pkgb --name=pkgb --version=1.0")
        c.run("install --requires=pkga/1.0 --requires=pkgb/1.0 --deployer=runtime_deploy "
              "--deployer-folder=myruntime -vvv")

        expected = sorted(["pkga.so", "pkgb.so", "pkga.dll"])
        assert sorted(os.listdir(os.path.join(c.current_folder, "myruntime"))) == expected

    def test_runtime_not_deploy(self):
        # https://github.com/conan-io/conan/issues/16712
        # If no run=False (no package-type), then no runtime is deployed
        c = TestClient()
        conanfile = textwrap.dedent("""
           from conan import ConanFile
           from conan.tools.files import copy
           class Pkg(ConanFile):
               def package(self):
                   copy(self, "*.so", src=self.build_folder, dst=self.package_folder)
                   copy(self, "*.dll", src=self.build_folder, dst=self.package_folder)
           """)
        c.save({"pkga/conanfile.py": conanfile,
                "pkga/lib/pkga.so": "",
                "pkga/bin/pkga.dll": ""})
        c.run("export-pkg pkga --name=pkga --version=1.0")
        c.run("install --requires=pkga/1.0 --deployer=runtime_deploy --deployer-folder=myruntime")
        assert os.listdir(os.path.join(c.current_folder, "myruntime")) == []

    def test_runtime_deploy_components(self):
        c = TestClient()
        conanfile = textwrap.dedent("""
            import os
            from conan import ConanFile
            from conan.tools.files import copy
            class Pkg(ConanFile):
               package_type = "shared-library"
               def package(self):
                   copy(self, "*.so", src=self.build_folder,
                        dst=os.path.join(self.package_folder, "a"))
                   copy(self, "*.dll", src=self.build_folder,
                        dst=os.path.join(self.package_folder, "b"))
               def package_info(self):
                   self.cpp_info.components["a"].libdirs = ["a"]
                   self.cpp_info.components["b"].bindirs = ["b"]
           """)
        c.save({"pkga/conanfile.py": conanfile,
                "pkga/lib/pkga.so": "",
                "pkga/bin/pkga.dll": "",
                "pkgb/conanfile.py": conanfile,
                "pkgb/lib/pkgb.so": ""})
        c.run("export-pkg pkga --name=pkga --version=1.0")
        c.run("export-pkg pkgb --name=pkgb --version=1.0")
        c.run("install --requires=pkga/1.0 --requires=pkgb/1.0 --deployer=runtime_deploy "
              "--deployer-folder=myruntime -vvv")

        assert sorted(os.listdir(os.path.join(c.current_folder, "myruntime"))) == sorted(["bin", "lib"])
        assert sorted(os.listdir(os.path.join(c.current_folder, "myruntime", "lib"))) == sorted(['pkga.so', 'pkgb.so'])
        assert sorted(os.listdir(os.path.join(c.current_folder, "myruntime", "bin"))) == sorted(['pkga.dll'])


@pytest.mark.parametrize("symlink, expected",
                         [(True, ["libfoo.so.0.1.0", "libfoo.so.0", "libfoo.so"]),
                          (False, ["libfoo.so.0.1.0"])])
def test_runtime_deploy_symlinks(symlink, expected):
    """ The deployer runtime_deploy should preserve symlinks when deploying shared libraries
    """
    c = TestClient()
    conanfile = textwrap.dedent("""
           from conan import ConanFile
           from conan.tools.files import copy, chdir
           import os
           class Pkg(ConanFile):
               package_type = "shared-library"
               def package(self):
                   copy(self, "*.so*", src=self.build_folder, dst=self.package_folder)
                   with chdir(self, os.path.join(self.package_folder, "lib")):
                       os.symlink(src="libfoo.so.0.1.0", dst="libfoo.so.0")
                       os.symlink(src="libfoo.so.0", dst="libfoo.so")
           """)
    c.save({"foo/conanfile.py": conanfile,
            "foo/lib/libfoo.so.0.1.0": ""})
    c.run("export-pkg foo/ --name=foo --version=0.1.0")
    c.run(f"install --requires=foo/0.1.0 --deployer=runtime_deploy --deployer-folder=output -c:a tools.deployer:symlinks={symlink}")

    sorted_expected = sorted(expected)
    assert sorted(os.listdir(os.path.join(c.current_folder, "output"))) == sorted_expected
    link_so_0 = os.path.join(c.current_folder, "output", "libfoo.so.0")
    link_so = os.path.join(c.current_folder, "output", "libfoo.so")
    lib = os.path.join(c.current_folder, "output", "libfoo.so.0.1.0")
    # INFO: This test requires in Windows to have symlinks enabled, otherwise it will fail
    if symlink and platform.system() != "Windows":
        assert os.path.islink(link_so_0)
        assert os.path.islink(link_so)
        assert not os.path.isabs(os.readlink(link_so_0))
        assert not os.path.isabs(os.readlink(os.path.join(link_so)))
        assert os.path.realpath(link_so) == os.path.realpath(link_so_0)
        assert os.path.realpath(link_so_0) == os.path.realpath(lib)
        assert not os.path.islink(lib)
    else:
        assert not os.path.islink(lib)


def test_runtime_deploy_subfolder():
    """ The deployer runtime_deploy should preserve subfolder structure when deploying shared libraries
    """
    c = TestClient()
    conanfile = textwrap.dedent("""
           from conan import ConanFile
           from conan.tools.files import copy
           import os
           class Pkg(ConanFile):
               package_type = "shared-library"
               def package(self):
                   copy(self, "*.so*", src=self.build_folder, dst=self.package_folder, keep_path=True)
           """)
    c.save({"foo/conanfile.py": conanfile,
            "foo/lib/libfoo.so": "",
            "foo/lib/subfolder/libbar.so": "",
            "foo/lib/subfolder/subsubfolder/libqux.so": "",})
    c.run("export-pkg foo/ --name=foo --version=0.1.0")
    c.run(f"install --requires=foo/0.1.0 --deployer=runtime_deploy --deployer-folder=output")

    assert sorted(os.listdir(os.path.join(c.current_folder, "output"))) == ["libfoo.so", "subfolder"]
    assert sorted(os.listdir(os.path.join(c.current_folder, "output", "subfolder"))) == ["libbar.so", "subsubfolder"]
    assert sorted(os.listdir(os.path.join(c.current_folder, "output", "subfolder", "subsubfolder"))) == ["libqux.so"]


def test_runtime_deploy_subfolder_symlink():
    """ The deployer runtime_deploy should preserve subfolder structure when deploying shared
        libraries with symlinks
    """
    c = TestClient()
    conanfile = textwrap.dedent("""
           from conan import ConanFile
           from conan.tools.files import copy, chdir, mkdir
           import os
           class Pkg(ConanFile):
               package_type = "shared-library"
               def package(self):
                   copy(self, "*.so*", src=self.build_folder, dst=self.package_folder)
                   with chdir(self, os.path.join(self.package_folder, "lib")):
                        os.symlink(src="subfolder/libfoo.so.1.0", dst="libfoo.so.1")
                        os.symlink(src="libfoo.so.1", dst="libfoo.so")
           """)
    c.save({"foo/conanfile.py": conanfile,
            "foo/lib/subfolder/libfoo.so.1.0": "",})
    c.run("export-pkg foo/ --name=foo --version=0.1.0")
    c.run(f"install --requires=foo/0.1.0 --deployer=runtime_deploy --deployer-folder=output -c:a tools.deployer:symlinks=True")

    assert os.listdir(os.path.join(c.current_folder, "output", "subfolder")) == ["libfoo.so.1.0"]
    # INFO: This test requires in Windows to have symlinks enabled, otherwise it will fail
    if platform.system() != "Windows":
        assert sorted(os.listdir(os.path.join(c.current_folder, "output"))) == ["libfoo.so", "libfoo.so.1" , "subfolder"]
        link_so_0 = os.path.join(c.current_folder, "output", "libfoo.so.1")
        link_so = os.path.join(c.current_folder, "output", "libfoo.so")
        lib = os.path.join(c.current_folder, "output", "subfolder", "libfoo.so.1.0")
        assert os.path.islink(link_so_0)
        assert os.path.islink(link_so)
        assert not os.path.isabs(os.readlink(link_so_0))
        assert not os.path.isabs(os.readlink(os.path.join(link_so)))
        assert os.path.realpath(link_so) == os.path.realpath(link_so_0)
        assert os.path.realpath(link_so_0) == os.path.realpath(lib)


def test_deployer_errors():
    c = TestClient()
    c.save({"conanfile.txt": "",
            "mydeploy.py": "",
            "mydeploy2.py": "nonsense"})
    c.run("install . --deployer=nonexisting.py", assert_error=True)
    assert "ERROR: Cannot find deployer 'nonexisting.py'" in c.out
    c.run("install . --deployer=mydeploy.py", assert_error=True)
    assert "ERROR: Deployer does not contain 'deploy()' function" in c.out
    c.run("install . --deployer=mydeploy2.py", assert_error=True)
    # The error message says conanfile, not a big deal still path to file is shown
    assert "ERROR: Unable to load conanfile" in c.out


def test_deploy_relative_paths():
    c = TestClient()
    consumer = textwrap.dedent("""
        import os
        from conan import ConanFile
        class Consumer(ConanFile):
            requires = "pkg/0.1"
            settings = "build_type"
            os = "build_type"
            generators = "CMakeDeps"
            def layout(self):
                self.folders.build = os.path.join("some/sub/folders")
                self.folders.generators = os.path.join(self.folders.build, "generators")
        """)
    deploy = textwrap.dedent("""
        import os, shutil
        def deploy(graph, output_folder):
            conanfile = graph.root.conanfile
            output_folder = os.path.join(conanfile.build_folder, "installed")
            for dep in conanfile.dependencies.values():
                new_folder = os.path.join(output_folder, dep.ref.name)
                shutil.copytree(dep.package_folder, new_folder, symlinks=True)
                dep.set_deploy_folder(new_folder)
        """)
    c.save({"pkg/conanfile.py": GenConanfile("pkg", "0.1"),
            "consumer/conanfile.py": consumer,
            "mydeploy.py": deploy})
    c.run("create pkg")

    # If we don't change to another folder, the full_deploy will be recursive and fail
    c.run("install consumer --build=missing --deployer=mydeploy.py")
    data = c.load("consumer/some/sub/folders/generators/pkg-release-data.cmake")
    assert 'set(pkg_PACKAGE_FOLDER_RELEASE "${CMAKE_CURRENT_LIST_DIR}/../installed/pkg")' in data


@pytest.mark.parametrize("absolute_path", [True, False])
def test_deploy_output_absolute(absolute_path):
    # https://github.com/conan-io/conan/issues/18560
    c = TestClient()
    c.save({"pkg/conanfile.py": GenConanfile("pkg", "0.1").with_package_file("myfile.txt", "c"),
            "consumer/conanfile.txt": "[requires]\npkg/0.1"})
    c.run("create pkg")

    # It is important to use the forward /myout in Windows, this was the breaking input
    out_path = f"{c.current_folder}/myout" if absolute_path else "myout"
    c.run("install consumer/conanfile.txt -s arch=x86_64 --deployer=full_deploy -g CMakeDeps "
          f'-of="{out_path}"')
    assert c.load("myout/full_deploy/host/pkg/0.1/myfile.txt") == "c"
    data = c.load("myout/pkg-release-x86_64-data.cmake")
    assert ('set(pkg_PACKAGE_FOLDER_RELEASE '
            '"${CMAKE_CURRENT_LIST_DIR}/full_deploy/host/pkg/0.1")') in data



import os
import platform
import textwrap

import pytest

from conan.test.utils.tools import TestClient
from conan.internal.util.files import rmdir


@pytest.mark.skipif(platform.system() != "Darwin", reason="Only OSX")
@pytest.mark.tool("cmake", "3.23")
def test_create_universal_binary():
    client = TestClient()

    conanfile = textwrap.dedent("""
        from conan import ConanFile
        from conan.tools.cmake import CMake, cmake_layout
        class mylibraryRecipe(ConanFile):
            package_type = "library"
            generators = "CMakeToolchain"
            settings = "os", "compiler", "build_type", "arch"
            options = {"shared": [True, False], "fPIC": [True, False]}
            default_options = {"shared": False, "fPIC": True}
            exports_sources = "CMakeLists.txt", "src/*", "include/*"

            def layout(self):
                cmake_layout(self)

            def build(self):
                cmake = CMake(self)
                cmake.configure()
                cmake.build()
                self.run("lipo -info libmylibrary.a")

            def package(self):
                cmake = CMake(self)
                cmake.install()

            def package_info(self):
                self.cpp_info.libs = ["mylibrary"]
    """)

    test_conanfile = textwrap.dedent("""
        import os
        from conan import ConanFile
        from conan.tools.cmake import CMake, cmake_layout
        from conan.tools.build import can_run

        class mylibraryTestConan(ConanFile):
            settings = "os", "compiler", "build_type", "arch"
            generators = "CMakeDeps", "CMakeToolchain"

            def requirements(self):
                self.requires(self.tested_reference_str)

            def build(self):
                cmake = CMake(self)
                cmake.configure()
                cmake.build()

            def layout(self):
                cmake_layout(self)

            def test(self):
                exe = os.path.join(self.cpp.build.bindir, "example")
                self.run(f"lipo {exe} -info", env="conanrun")
            """)

    client.run("new cmake_lib -d name=mylibrary -d version=1.0")
    client.save({"conanfile.py": conanfile, "test_package/conanfile.py": test_conanfile})

    client.run('create . --name=mylibrary --version=1.0 '
               '-s="arch=armv8|armv8.3|x86_64" --build=missing -tf=""')

    assert "libmylibrary.a are: x86_64 arm64 arm64e" in client.out

    client.run('test test_package mylibrary/1.0 -s="arch=armv8|armv8.3|x86_64"')

    assert "example are: x86_64 arm64 arm64e" in client.out

    client.run('new cmake_exe -d name=foo -d version=1.0 -d requires=mylibrary/1.0 --force')

    client.run('install . -s="arch=armv8|armv8.3|x86_64"')

    client.run_command("cmake --preset conan-release")
    client.run_command("cmake --build --preset conan-release")
    client.run_command("lipo -info ./build/Release/foo")

    assert "foo are: x86_64 arm64 arm64e" in client.out

    rmdir(os.path.join(client.current_folder, "build"))

    client.run('install . -s="arch=armv8|armv8.3|x86_64" '
               '-c tools.cmake.cmake_layout:build_folder_vars=\'["settings.arch"]\'')

    client.run_command("cmake --preset \"conan-armv8_armv8.3_x86_64-release\" ")
    client.run_command("cmake --build --preset \"conan-armv8_armv8.3_x86_64-release\" ")
    client.run_command("lipo -info './build/armv8_armv8.3_x86_64/Release/foo'")

    assert "foo are: x86_64 arm64 arm64e" in client.out


@pytest.mark.skipif(platform.system() != "Darwin", reason="Only OSX")
@pytest.mark.tool("cmake", "3.23")
def test_create_universal_binary_ninja():
    client = TestClient()

    client.run("new cmake_lib -d name=mylibrary -d version=1.0")

    client.run('export .')

    conanfile = textwrap.dedent("""
        [requires]
        mylibrary/1.0
        [generators]
        CMakeToolchain
        CMakeDeps
        VirtualBuildEnv
        VirtualRunEnv
        """)

    cmake = textwrap.dedent("""
        cmake_minimum_required(VERSION 3.23)
        project(ninjatest NONE)

        find_package(mylibrary CONFIG REQUIRED)
        """)

    client.save({"conanfile.txt": conanfile,
                 "CMakeLists.txt": cmake},
                clean_first=True)

    client.run('install . -s=arch="armv8|x86_64" --build=missing -of=build')

    with client.chdir("build"):
        client.run_command("cmake .. -GNinja "
                           "-DCMAKE_TOOLCHAIN_FILE=conan_toolchain.cmake "
                           "-DCMAKE_POLICY_DEFAULT_CMP0091=NEW "
                           "-DCMAKE_BUILD_TYPE=Release")

    assert "expected newline, got '|'" not in client.out
    assert "Build files have been written to:" in client.out
    # test that there are no files with the "|" character in the build folder
    assert not any("|" in f for f in os.listdir(os.path.join(client.current_folder, "build")))



import os
import sys

from conan.api.output import init_colorama
from conan.api.subapi.audit import AuditAPI
from conan.api.subapi.cache import CacheAPI
from conan.api.subapi.command import CommandAPI
from conan.api.subapi.local import LocalAPI
from conan.api.subapi.lockfile import LockfileAPI
from conan.api.subapi.report import ReportAPI
from conan.api.subapi.workspace import WorkspaceAPI
from conan.api.subapi.config import ConfigAPI
from conan.api.subapi.download import DownloadAPI
from conan.api.subapi.export import ExportAPI
from conan.api.subapi.install import InstallAPI
from conan.api.subapi.graph import GraphAPI
from conan.api.subapi.new import NewAPI
from conan.api.subapi.profiles import ProfilesAPI
from conan.api.subapi.list import ListAPI
from conan.api.subapi.remotes import RemotesAPI
from conan.api.subapi.remove import RemoveAPI
from conan.api.subapi.upload import UploadAPI
from conan.errors import ConanException
from conan.internal.cache.home_paths import HomePaths
from conan.internal.hook_manager import HookManager
from conan.internal.model.conf import load_global_conf, ConfDefinition, CORE_CONF_PATTERN
from conan.internal.model.settings import load_settings_yml
from conan.internal.paths import get_conan_user_home
from conan.internal.api.migrations import ClientMigrator
from conan.internal.model.version_range import validate_conan_version
from conan.internal.rest.conan_requester import ConanRequester


class ConanAPI:
    """
    This is the main object to interact with the Conan API. It provides all the subapis to work with
    recipes, packages, remotes, etc., which are exposed as attributes of this class, and should
    not be created directly.
    """
    def __init__(self, cache_folder=None):
        """
        :param cache_folder: Conan cache/home folder. It will have less priority than the
                             ``"home_folder"`` defined in a Workspace.
        """

        version = sys.version_info
        if version.major == 2 or version.minor < 7:
            raise ConanException("Conan needs Python >= 3.7")
        if cache_folder is not None and not os.path.isabs(cache_folder):
            raise ConanException("cache_folder has to be an absolute path")

        init_colorama(sys.stderr)
        # Deprecated, but still used internally, prefer home_folder
        self.cache_folder = cache_folder or get_conan_user_home()
        self._home_folder = self.cache_folder
        self._api_helpers = self._ApiHelpers(self)
        self.migrate()

        #: Used to interact with the local Conan configuration
        self.config: ConfigAPI = ConfigAPI(self, self._api_helpers)
        #: Used to interact with remotes
        self.remotes: RemotesAPI = RemotesAPI(self, self._api_helpers)
        self.command = CommandAPI(self)
        #: Used to get latest refs and list refs of recipes and packages
        self.list: ListAPI = ListAPI(self)
        self.profiles = ProfilesAPI(self, self._api_helpers)
        self.install = InstallAPI(self, self._api_helpers)
        self.graph = GraphAPI(self, self._api_helpers)
        #: Used to export recipes and pre-compiled package binaries to the Conan cache
        self.export: ExportAPI = ExportAPI(self, self._api_helpers)
        self.remove = RemoveAPI(self)
        self.new = NewAPI(self)
        #: Used to upload recipes and packages to remotes
        self.upload: UploadAPI = UploadAPI(self, self._api_helpers)
        #: Used to download recipes and packages from remotes
        self.download: DownloadAPI = DownloadAPI(self)
        #: Used to interact wit the packages storage cache
        self.cache: CacheAPI = CacheAPI(self, self._api_helpers)
        #: Used to read and manage lockfile files
        self.lockfile: LockfileAPI = LockfileAPI(self)
        self.local = LocalAPI(self, self._api_helpers)
        #: Used to check vulnerabilities of dependencies
        self.audit: AuditAPI = AuditAPI(self)
        # Now, lazy loading of editables
        self.workspace = WorkspaceAPI(self)
        self.report: ReportAPI = ReportAPI(self, self._api_helpers)

    @property
    def home_folder(self) -> str:
        """ Where the Conan user home is located. Read only.
        Can be modified by the ``CONAN_HOME`` environment variable or by the
        ``.conanrc`` file in the current directory or any parent directory
        when Conan is called.
        """
        return self._home_folder

    def reinit(self):
        """
        Reinitialize the Conan API. This is useful when the configuration changes.
        """
        self._api_helpers.reinit()
        self.local.reinit()

    def migrate(self):
        # Migration system
        # TODO: A prettier refactoring of migrators would be nice
        from conan import conan_version
        migrator = ClientMigrator(self.cache_folder, conan_version)
        migrator.migrate()

    class _ApiHelpers:
        def __init__(self, conan_api):
            self._conan_api = conan_api
            self._cli_core_confs = None
            self._init_global_conf()
            self.hook_manager = HookManager(HomePaths(self._conan_api.home_folder).hooks_path)
            # Wraps an http_requester to inject proxies, certs, etc
            self._requester = ConanRequester(self.global_conf, self._conan_api.home_folder)
            self._settings_yml = None

        def set_core_confs(self, core_confs):
            confs = ConfDefinition()
            for c in core_confs:
                if not CORE_CONF_PATTERN.match(c):
                    raise ConanException(f"Only core. values are allowed in --core-conf. Got {c}")
            confs.loads("\n".join(core_confs))
            confs.validate()
            self._cli_core_confs = confs
            # Last but not least, apply the new configuration
            # This will in turn call ApiHelpers.reinit() as the very first thing
            self._conan_api.reinit()

        def _init_global_conf(self):
            self.global_conf = load_global_conf(self._conan_api.home_folder)
            if self._cli_core_confs:
                self.global_conf.update_conf_definition(self._cli_core_confs)
            required_range_new = self.global_conf.get("core:required_conan_version")
            if required_range_new:
                validate_conan_version(required_range_new)

        def reinit(self):
            self._init_global_conf()
            self.hook_manager.reinit()
            self._requester = ConanRequester(self.global_conf, self._conan_api.home_folder)
            self._settings_yml = None

        @property
        def settings_yml(self):
            if self._settings_yml is None:
                self._settings_yml = load_settings_yml(self._conan_api.home_folder)
            return self._settings_yml

        @property
        def requester(self):
            return self._requester



"""
This file is able to create a self contained Conan executable that contains all it needs,
including the Python interpreter, so it wouldnt be necessary to have Python installed
in the system
It is important to install the dependencies and the project first with "pip install -e ."
which configures the project as "editable", that is, to run from the current source folder
After creating the executable, it can be pip uninstalled

$ pip install -e .
$ python pyinstaller.py

This has to run in the same platform that will be using the executable, pyinstaller does
not cross-build

The resulting executable can be put in the system PATH of the running machine
"""

import argparse
import os
import platform
import shutil
import sys

from conan import __version__


def _run_bin(pyinstaller_path):
    # run the binary to test if working
    conan_bin = os.path.join(pyinstaller_path, 'dist', 'conan', 'conan')
    if platform.system() == 'Windows':
        conan_bin = '"' + conan_bin + '.exe' + '"'
    retcode = os.system(conan_bin)
    if retcode != 0:
        raise Exception("Binary not working")


def _windows_version_file(version):
    template = """# UTF-8
#
# For more details about fixed file info 'ffi' see:
# http://msdn.microsoft.com/en-us/library/ms646997.aspx
VSVersionInfo(
  ffi=FixedFileInfo(
    # filevers and prodvers should be always a tuple with four items: (1, 2, 3, 4)
    # Set not needed items to zero 0.
    filevers={version_tuple},
    prodvers={version_tuple},
    # Contains a bitmask that specifies the valid bits 'flags'r
    mask=0x3f,
    # Contains a bitmask that specifies the Boolean attributes of the file.
    flags=0x0,
    # The operating system for which this file was designed.
    # 0x4 - NT and there is no need to change it.
    OS=0x4,
    # The general type of file.
    # 0x1 - the file is an application.
    fileType=0x1,
    # The function of the file.
    # 0x0 - the function is not defined for this fileType
    subtype=0x0,
    # Creation date and time stamp.
    date=(0, 0)
    ),
  kids=[
    StringFileInfo(
      [
      StringTable(
        u'000004b0',
        [StringStruct(u'Comments', u'This executable was created with pyinstaller'),
        StringStruct(u'CompanyName', u'JFrog'),
        StringStruct(u'FileDescription', u'Conan C, C++ Open Source Package Manager'),
        StringStruct(u'FileVersion', u'{version}'),
        StringStruct(u'LegalCopyright', u'Copyright 2020 JFrog'),
        StringStruct(u'ProductName', u'Conan'),
        StringStruct(u'ProductVersion', u'{version}')])
      ]),
    VarFileInfo([VarStruct(u'Translation', [0, 1200])])
  ]
)"""
    if "-" in version:
        version, _ = version.split("-")
    version_tuple = tuple([int(v) for v in version.split(".")] + [0])
    return template.format(version=version, version_tuple=version_tuple)


def pyinstall(source_folder, onefile=False):
    try:
        import PyInstaller.__main__  # noqa  Pyinstaller is not in the default requirements.txt
    except ImportError:
        print(f"ERROR: PyInstaller not found. Please install it with "
              f"'python -m pip install pyinstaller'")
        sys.exit(-1)

    pyinstaller_path = os.path.join(os.getcwd(), 'pyinstaller')
    if not os.path.exists(pyinstaller_path):
        os.mkdir(pyinstaller_path)
    try:
        shutil.rmtree(os.path.join(pyinstaller_path))
    except Exception as e:
        print("Unable to remove old folder", e)

    conan_path = os.path.join(source_folder, 'conans', 'conan.py')
    hidden_imports = []

    if not os.path.exists(pyinstaller_path):
        os.mkdir(pyinstaller_path)

    if platform.system() != "Windows":
        hidden_imports.append("setuptools.msvc")
        win_ver = ""
    else:
        win_ver_file = os.path.join(pyinstaller_path, 'windows-version-file')
        content = _windows_version_file(__version__)
        with open(win_ver_file, 'w') as file:
            file.write(content)
        win_ver = ["--version-file", win_ver_file]

    if onefile:
        distpath = os.path.join(pyinstaller_path, "dist", "conan")
    else:
        distpath = os.path.join(pyinstaller_path, "dist")

    command_args = [conan_path, "--noconfirm", f"--paths={source_folder}",
                    "--console", f"--distpath={distpath}"]

    # Python standard library modules to ensure in the bundle
    hidden_imports.append("glob")
    hidden_imports.append("pathlib")
    # inclusion of full conan package (with exclusion of irrelevant modules)
    command_args.append("--collect-submodules=conan")
    command_args.append("--exclude-module=conan.test.*")

    command_args.extend(f"--hidden-import={name}" for name in hidden_imports)

    if win_ver:
        command_args.extend(win_ver)
    if onefile:
        command_args.append("--onefile")

    PyInstaller.__main__.run(command_args)

    _run_bin(pyinstaller_path)

    return os.path.abspath(os.path.join(pyinstaller_path, 'dist', 'conan'))


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument('--onefile', action='store_true')
    parser.set_defaults(onefile=False)
    args = parser.parse_args()

    src_folder = os.path.abspath(os.path.dirname(os.path.abspath(__file__)))
    output_folder = pyinstall(src_folder, args.onefile)
    print("\n**************Conan binaries created!******************\n"
          "\nAppend this folder to your system PATH: '%s'\n"
          "Feel free to move the whole folder to another location." % output_folder)
